"""A few fundamental tools to operate on a fixed collection of objects (e.g. functions).

For functions you have:

- ``Pipe``: To compose functions (output of one fed as the input of the next)

- ``FuncFanout``: To apply multiple functions to the same inputs

- ``FlexFuncFanout``: Like `FuncFanout` but where the application of inputs is flexible.
That is, the functions "draw" their inputs from the a common pool, but don't choke
if there are extra unrecognized arguments.

- ``ParallelFuncs``: To make a dict-to-dict function, applying a specific function for
each input key (putting the result in that key in the output.

For context managers you have:

- ``ContextFanout``: To hold multiple context managers as one (entering and exiting
together)

![image](https://user-images.githubusercontent.com/1906276/138004878-bfe17115-c25f-4d22-9740-0fef983507c0.png)

"""
from typing import Mapping, Iterable, Union, Callable, Any, TypeVar
from inspect import Signature, signature


def name_of_obj(o: object) -> Union[str, None]:
    """
    Tries to find the (or "a") name for an object, even if `__name__` doesn't exist.

    >>> name_of_obj(map)
    'map'
    >>> name_of_obj([1, 2, 3])
    'list'
    >>> name_of_obj(print)
    'print'
    >>> name_of_obj(lambda x: x)
    '<lambda>'
    >>> from functools import partial
    >>> name_of_obj(partial(print, sep=","))
    'print'
    """
    if hasattr(o, '__name__'):
        return o.__name__
    elif hasattr(o, '__class__'):
        name = name_of_obj(o.__class__)
        if name == 'partial':
            if hasattr(o, 'func'):
                return name_of_obj(o.func)
        return name
    else:
        return None


def ensure_iterable_of_callables(x):
    """Assert that the input is an iterable of callables,
    or wrap a single callable in an iterable.
    """
    if isinstance(x, Iterable):
        if not all(callable(xx) for xx in x):
            non_callables = filter(lambda xx: not callable(xx), x)
            raise TypeError(f'These were not callable: {list(non_callables)}')
        return x
    else:
        assert callable(x)
        return (x,)


Obj = TypeVar('Obj')


def uniquely_named_objects(
    objects: Iterable[Obj],
    exclude_names: Iterable[str] = (),
    obj_to_name: Callable[[Obj], str] = name_of_obj,
    *,
    name_for_position: dict = (),
):
    """Generate (name, object) pairs from an iterable of objects

    :param objects: Objects to be named
    :param exclude_names: Names that can't be used
    :param obj_to_name: Function that tries to get/make a name from an object
    :param name_for_position: A ``{position_idx: name,...}`` mapping that instructs
    ``uniquely_named_objects`` to use a specific name for a given position.

    >>> from functools import partial
    >>> objects = [map, [1], [1, 2], lambda x: x, partial(print, sep=",")]
    >>> g = uniquely_named_objects(objects)
    >>> names_and_objects = dict(g)
    >>> list(names_and_objects)
    ['map', 'list', '_2', '_3', 'print']

    That ``'_2'`` is there because both ``[1]`` and ``[1, 2]`` would be named `'list'`,
    so to avoid that, a default name (revealing the position of the object in the
    input `objects`) is given.
    The '_3' comes from the fact that the ``lambda`` function doesn't have a proper
    name (one that is a python identifier).

    If we wanted the name for ``[1]`` to revert to the default positional name ``'_1'``,
    we can achieve this by forbidding the name ``'list'``:

    >>> list(dict(uniquely_named_objects(objects, exclude_names={'list'})))
    ['map', '_1', '_2', '_3', 'print']

    You could also acheive this by specifying this ``'_1'`` explicitly in the
    ``name_for_position`` argument:

    >>> list(dict(uniquely_named_objects(objects, name_for_position={1: '_1'})))
    ['map', '_1', 'list', '_3', 'print']

    The reason this `list` reappears as a name is that we didn't exclude it, and
    the name is not taken by the ``[1]`` argument anymore.
    To get the desired effect with ``name_for_position`` we could therefore do this:

    >>> list(dict(uniquely_named_objects(objects, name_for_position={1: '_1', 2: '_2'})))
    ['map', '_1', '_2', '_3', 'print']

    Obviously, ``exclude_names`` is the right argument for the problem above, but
    what ``name_for_position`` does give you is the ability to explicitly chose the
    names you want to assign to all or some of the elements of your iterable.

    >>> list(dict(uniquely_named_objects(
    ...     objects, name_for_position={1: 'first_list', 2: 'second_list', 3: 'lambda'}))
    ... )
    ['map', 'first_list', 'second_list', 'lambda', 'print']

    Extra notes:

    See what ``uniquely_named_objects`` offers as parametrization:

    - You can provide an exclusion list (though the handing of a conflict is hardcoded
    and questionable)

    - You can provide a ``obj_to_name`` function to control the naming of objects.
    One trick to be aware of if objects have unique hashes: Make a
    ``d = {obj: name,...}`` mapping and specify ``obj_to_name=d.get``.

    - Any controllable way to decide on a name based on the position of the function
    in the iterable (this could be useful!)

    What you DO NOT have:

    - Any way to choose names non-myopically: An object’s name cannot “see” the
    objects around it to decide on a name (it can only see the names use by those behind
    it through ``exclude_names``).

    - Any “retries” or “alternative naming logic” if a chosen name conflicts with
    ``exclude_names``
    """
    _exclude_names = set(exclude_names)
    for i, obj in enumerate(objects):
        if i in name_for_position:
            name = name_for_position[i]
        else:
            name = obj_to_name(obj)
        # if name == '<lambda>':
        # name = f'lambda_{i}'
        if name is None or not name.isidentifier() or name in _exclude_names:
            name = f'_{i}'
            assert (
                name not in _exclude_names
            ), '{name} already used in {_exclude_names}!'
        yield name, obj
        _exclude_names.add(name)


def merge_unnamed_and_named(*unnamed, **named) -> dict:
    """To merge unnamed and named arguments into a single (named) dict of arguments

    >>> merge_unnamed_and_named(10, 20, thirty=30, fourty=40)
    {'_0': 10, '_1': 20, 'thirty': 30, 'fourty': 40}
    """
    # TODO: Could do what is done in meshed (try to get names of actual functions)
    #  and use _0, _1, _2... as fallback only?
    named_unnamed = {f'_{i}': obj for i, obj in enumerate(unnamed)}
    if not named_unnamed.keys().isdisjoint(named):
        raise ValueError(
            f"Some of your objects' names clashed: "
            f'{named_unnamed.keys() & named.keys()}'
        )
    return dict(named_unnamed, **named)


# TODO: Refactoring: Remove once not used anymore
def _multi_func_init(self, *unnamed_funcs, **named_funcs) -> None:
    if len(unnamed_funcs) == 1 and isinstance(unnamed_funcs[0], Mapping):
        self.funcs = unnamed_funcs[0]
        expected_n_funcs = len(self.funcs)
    else:
        expected_n_funcs = len(unnamed_funcs) + len(named_funcs)
        self.funcs = merge_unnamed_and_named(*unnamed_funcs, **named_funcs)
    if len(self.funcs) != expected_n_funcs:
        raise ValueError(
            'Some of your func names clashed. Your unnamed funcs were: '
            f'{unnamed_funcs} and your named ones were: {named_funcs}'
        )
    ensure_iterable_of_callables(self.funcs.values())


_dflt_signature = Signature.from_callable(lambda *args, **kwargs: None)


class MultiObj(Mapping):
    """A base class that holds several named objects

    >>> from functools import partial

    Let's make a `MultiObj` with some miscellaneous objects.
    (Note that `MultiObj` will usually be used for specific kinds of objects such as
    callables or context managers. Here we chose the objects to demo what the
    auto-naming does.)

    >>> mo = MultiObj([1], [1, 2], partial(print, sep=","), i='hi', ident=lambda x: x)

    You now have a mapping and can do mapping things such as being able to list keys,
    getting the length, seeing if a key is present, and getting the value for a key.

    Note that the first and second list cannot be assigned the same name without creating a conflict.
    In general one of the following will happen:
    -you give it a name
    -it tries to figure out a non conflicting name (if the object has a dunder name, etc)
    -it falls back to a naming that is just the stringification of the argument's positional index

    >>> list(mo) # not that the second item cannot be 'list', so a different name is given
    ['list', '_1', 'print', 'i', 'ident']
    >>> len(mo)
    5
    >>> mo['_1']
    [1, 2]
    >>> 'list' in mo
    True
    >>> 'not a key of mo' in mo
    False

    When a key (always a string) is also a valid identifier, and in-so-far as
    it doesn't clash with other attributes, `MultiObj` will also give
    you access to the names/keys of your objects via attributes.
    (Note, this is similar to what `pandas.DataFrame` does with it's columns names.)

    >>> mo.list
    [1]
    >>> mo.print
    functools.partial(<built-in function print>, sep=',')

    You can also specify an object mapping directly through a mapping:

    >>> mo = MultiObj({'this': [1], 'that': [1, 2]})
    >>> dict(mo)
    {'this': [1], 'that': [1, 2]}

    You can specify an instance name and/or doc with the special (reserved) argument
    names ``__name__`` and ``__doc__`` (which therefore can't be used as object names:

    >>> mo = MultiObj(
    ... this=[1], that=[1, 2], __name__='this_and_that', __doc__='Nothing much'
    ... )
    >>> dict(mo)
    {'this': [1], 'that': [1, 2]}
    >>> mo.__name__
    'this_and_that'
    >>> mo.__doc__
    'Nothing much'
    """

    # Placing as staticmethod so that subclasses can overwrite if necessary
    auto_namer = staticmethod(uniquely_named_objects)

    def __init__(self, *unnamed, **named):
        named = self._process_reserved_names(named)

        if len(unnamed) == 1 and isinstance(unnamed[0], Mapping) and len(named) == 0:
            # Special case where a single input is given: a mapping between names and obj
            self.objects = unnamed[0]
        else:
            # Normal case: Some named and unnamed objects are given,
            # so we need to get unique names for the unnamed
            named_unnamed = dict(self.auto_namer(unnamed, exclude_names=set(named)))
            # The uniquely_name_objects should take care of this, but extra precaution:
            if not named_unnamed.keys().isdisjoint(named):
                raise ValueError(
                    f"Some of your objects' names clashed: "
                    f'{named_unnamed.keys() & named.keys()}'
                )
            # Merge these
            self.objects = dict(named_unnamed, **named)

    _reserved_names = ('__name__', '__doc__')

    def _process_reserved_names(self, named_funcs):
        for name in self._reserved_names:
            if (value := named_funcs.pop(name, None)) is not None:
                setattr(self, name, value)
        return named_funcs

    def __iter__(self):
        yield from self.objects

    def __getitem__(self, k):
        return self.objects[k]

    def __contains__(self, k):
        return k in self.objects

    def __len__(self):
        return len(self.objects)

    def __getattr__(self, item):
        """Access to those objects that have proper identifier names"""
        if item in self and item.isidentifier():
            return self.objects[item]
        else:
            raise AttributeError(f'Not an attribute: {item}')

    def __repr__(self):
        return f"<{type(self).__name__} containing: {', '.join(self)}>"

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        for k, v in state.items():
            setattr(self, k, v)


def iterable_of_callables_validation(funcs: Iterable[Callable]):
    if not isinstance(funcs, Iterable):
        raise TypeError(f'Not an iterable: {funcs}')
    elif not all(callable(xx) for xx in funcs):
        non_callables = filter(lambda f: not callable(f), funcs)
        raise TypeError(f'These were not callable: {list(non_callables)}')


class MultiFunc(MultiObj):
    """A MultiObj, but specialized to contain callable objects only"""

    def __init__(self, *unnamed_funcs, **named_funcs):
        # The basic MultiObj initialization
        super().__init__(*unnamed_funcs, **named_funcs)
        # The extra stuff we want to do with functions
        iterable_of_callables_validation(self.funcs.values())

    @property
    def funcs(self):
        """Alias of .objects, for better readability in the context of MultiFunc"""
        return self.objects


def _signature_from_first_and_last_func(first_func, last_func):
    try:
        input_params = signature(first_func).parameters.values()
    except ValueError:  # function doesn't have a signature, so take default
        input_params = _dflt_signature.parameters.values()
    try:
        return_annotation = signature(last_func).return_annotation
    except ValueError:  # function doesn't have a signature, so take default
        return_annotation = _dflt_signature.return_annotation
    return Signature(input_params, return_annotation=return_annotation)


class Pipe(MultiFunc):
    """Simple function composition. That is, gives you a callable that implements

    input -> f_1 -> ... -> f_n -> output.

    >>> def foo(a, b=2):
    ...     return a + b
    >>> f = Pipe(foo, lambda x: print(f"x: {x}"))
    >>> f(3)
    x: 5

    You can name functions, but this would just be for documentation purposes.
    The names are completely ignored.

    >>> g = Pipe(
    ...     add_numbers = lambda x, y: x + y,
    ...     multiply_by_2 = lambda x: x * 2,
    ...     stringify = str
    ... )
    >>> g(2, 3)
    '10'

    Notes:
        - Pipe instances don't have a __name__ etc. So some expectations of normal functions are not met.
        - Pipe instance are pickalable (as long as the functions that compose them are)

    You can specify a single functions:

    >>> Pipe(lambda x: x + 1)(1)
    2

    but

    >>> Pipe()
    Traceback (most recent call last):
      ...
    ValueError: You need to specify at least one function!

    You can specify an instance name and/or doc with the special (reserved) argument
    names ``__name__`` and ``__doc__`` (which therefore can't be used as function names):

    >>> f = Pipe(map, add_it=sum, __name__='map_and_sum', __doc__='Apply func and add')
    >>> f(lambda x: x * 10, [1, 2, 3])
    60
    >>> f.__name__
    'map_and_sum'
    >>> f.__doc__
    'Apply func and add'
    """

    def __init__(self, *unnamed_funcs, **named_funcs):
        # The basic MultiFunc initialization
        super().__init__(*unnamed_funcs, **named_funcs)

        # The extra initialization for pipelines
        callables = list(self.funcs.values())
        n_funcs = len(callables)
        if n_funcs == 0:
            raise ValueError('You need to specify at least one function!')

        elif n_funcs == 1:
            other_funcs = ()
            first_func = last_func = callables[0]
        else:
            first_func, *other_funcs = callables
            *_, last_func = other_funcs

        self.__signature__ = _signature_from_first_and_last_func(first_func, last_func)
        self.first_func = first_func
        self.other_funcs = other_funcs

    def __call__(self, *args, **kwargs):
        out = self.first_func(*args, **kwargs)
        for func in self.other_funcs:
            out = func(out)
        return out

    def __len__(self):
        return len(self.funcs)

    def __eq__(self, other):
        return pipes_are_equal(self, other)


from operator import eq


def _flatten_pipe(pipe):
    for func in pipe.funcs.values():
        if isinstance(func, Pipe):
            yield from _flatten_pipe(func)
        else:
            yield func


def flatten_pipe(pipe):
    """Unravel nested Pipes to get a flat 'sequence of functions' version of input.

    >>> def f(x): return x + 1
    >>> def g(x): return x * 2
    >>> def h(x): return x - 3
    >>> a = Pipe(f, g, h)
    >>> b = Pipe(f, Pipe(g, h))
    >>> len(a)
    3
    >>> len(b)
    2
    >>> c = flatten_pipe(b)
    >>> len(c)
    3
    >>> assert a(10) == b(10) == c(10) == 19
    """
    return Pipe(*_flatten_pipe(pipe))


def pipes_are_equal(p1, p2, *, func_equality=eq, verbose=False):
    """Determine if two pipelines are equal.

    Pipelines are equal if their flattened versions have equal functions.
    Function equality can be controlled by the ``func_equality`` argument.
    The ``verbose`` argument will print some more information about why the pipelines
    are not equal.

    >>> def f(x): return x + 1
    >>> def g(x): return x * 2
    >>> def h(x): return x - 3
    >>> a = Pipe(f, g, h)
    >>> b = Pipe(f, g, h)
    >>> c = Pipe(f, Pipe(g, h))
    >>> assert a(10) == b(10) == c(10) == 19
    >>> pipes_are_equal(a, b)
    True
    >>> pipes_are_equal(a, c)
    True
    >>> pipes_are_equal(Pipe(f, g), Pipe(g, h))
    False
    >>> pipes_are_equal(Pipe(f, g), Pipe(f, g, h))
    False

    Get more information when pipes are not equal.

    >>> pipes_are_equal(Pipe(f, g), Pipe(f, g, h), verbose=True)
    --> Flattened pipes do not have the same number of functions: len(p1)=2 != len(p2)=3
    False

    Change how functions are compared for equality:

    >>> pipes_are_equal(Pipe(lambda x: x), Pipe(lambda x: x))
    False
    >>> from inspect import getsource
    >>> source_equality = lambda f, ff: getsource(f) == getsource(ff)
    >>> pipes_are_equal(
    ...     Pipe(lambda x: x), Pipe(lambda x: x), func_equality=source_equality
    ... )
    True

    """
    p1, p2 = map(flatten_pipe, (p1, p2))
    if len(p1) != len(p2):
        if verbose:
            print(
                f'--> Flattened pipes do not have the same number of functions:'
                f' {len(p1)=} != {len(p2)=}'
            )
        return False  # if not same number of keys or keys different, not equality
    else:
        for func1, func2 in zip(p1.funcs.values(), p2.funcs.values()):
            if not func_equality(func1, func2):
                if verbose:
                    print(f'--> func1 and func2 are not equal: {func1=} != {func2=}')
                return False  # these two funcs are not equal
            # else continue
    return True


class FuncFanout(MultiFunc):
    """Applies multiple functions to the same argument(s) and returns a dict of results.

    You know how `map(func, iterable_of_inputs)` applies a same function to an iterable
    of inputs.
    `FuncFanout` (we could call it `pam`) is a sort of dual; used to apply multiple
    functions to a same input.

    >>> def foo(a):
    ...     return a + 2
    ...
    >>> def bar(a):
    ...     return a * 2
    ...
    >>> groot = lambda a: 'I am groot'
    >>> m = FuncFanout(foo, bar, groot)
    >>>
    >>> list(m(3))
    [('foo', 5), ('bar', 6), ('_2', 'I am groot')]
    >>> dict(m(3))
    {'foo': 5, 'bar': 6, '_2': 'I am groot'}

    Don't like that `_2` name?
    Well, If you specify names to the input functions, they'll be used instead of the
    ones found by the `MultObj.auto_namer`.

    >>> m = FuncFanout(foo, bar_results=bar, groot=groot)
    >>> dict(m(10))
    {'foo': 12, 'bar_results': 20, 'groot': 'I am groot'}

    Or if you want your results as a tuple, you could do:

    >>> tuple(dict(m(10)).values())
    (12, 20, 'I am groot')

    The above, gather in a ``dict`` is one way to get your data, but what calling a
    ``FuncFanout`` instance actually gives you is a generator that yields the
    ``(func_key, func_output)`` pairs one at a time

    Sometimes you may want/need more control though, and prefer to iterate through the
    pairs yourself, and in that case use `call_generator` directly.

    >>> gen = m(10)
    >>> next(gen)
    ('foo', 12)
    >>> next(gen)
    ('bar_results', 20)
    >>> next(gen)
    ('groot', 'I am groot')

    So this gives you control on how you want your data.
    Here's a recipe: Say you want to make a function that gives you the data as a dict
    automatically. You can do this, using ``i2.Pipe``:

    >>> f = Pipe(m, dict)
    >>> f(10)
    {'foo': 12, 'bar_results': 20, 'groot': 'I am groot'}

    Or if you want a tuple:

    >>> from operator import itemgetter, methodcaller
    >>> from functools import partial
    >>> f = Pipe(m, partial(map, itemgetter(1)), tuple)
    >>> f(10)
    (12, 20, 'I am groot')

    """

    def call_generator(self, *args, **kwargs):
        for name, func in self.funcs.items():
            yield name, func(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.call_generator(*args, **kwargs)


from i2.signatures import ch_func_to_all_pk, tuple_the_args, Sig


# TODO: Finish this!
# TODO: Test the handling var positional and var keyword
class FlexFuncFanout(MultiFunc):
    """
    Call multiple functions, using a pool of arguments that they will draw from.

    >>> from i2.tests.objects_for_testing import formula1, sum_of_args, mult, add
    >>> mf1 = FlexFuncFanout(formula1=formula1, mult=mult, add=add)
    >>> kwargs_for_func = mf1.kwargs_for_func(w=1, x=2, z=3, a=4, b=5)

    What's this for? Well, the raison d'etre of `FlexFuncFanout` is to be able to do this:

    >>> assert add(a=4, b=5) == add(**kwargs_for_func['add'])

    This wouldn't work on all functions since some functions have position only arguments (e.g. ``formula1``).
    Therefore ``FlexFuncFanout`` holds a "normalized" form of the functions; namely one that handles such things as
    postion only and varargs.

    # TODO: Make this work!
    #   Right now raises: TypeError: formula1() got some positional-only arguments passed as keyword arguments: 'w'
    # >>> assert formula1(1, x=2, z=3) == mf1.normalized_funcs[formula1](**kwargs_for_func[formula1])

    Note: In the following, it looks like ``FlexFuncFanout`` instances return dicts whose keys are strings.
    This is not the case.
    The keys are functions: The same functions that were input.
    The reason for not using functions is that when printed, they include their hash, which invalidates the doctests.

    # >>> def print_dict(d):  # just a util for this doctest
    # ...     from pprint import pprint
    # ...     pprint({k.__name__: d[k] for k in sorted(d, key=lambda x: x.__name__)})

    >>> mf1 = FlexFuncFanout(formula1, mult=mult, addition=add)
    >>> assert mf1.kwargs_for_func(w=1, x=2, z=3, a=4, b=5) == {
    ... 'formula1': {'w': 1, 'x': 2, 'z': 3},
    ... 'mult': {'x': 2},
    ... 'addition': {'a': 4, 'b': 5},
    ... }

    Oh, and you can actually see the signature of kwargs_for_func:

    >>> from inspect import signature
    >>> signature(mf1)
    <Sig (w, x: float, a, y=1, z: int = 1, b: float = 0.0)>

    >>> mf2 = FlexFuncFanout(formula1, mult, addition=add, mysum=sum_of_args)
    >>> assert mf2.kwargs_for_func(
    ...     w=1, x=2, z=3, a=4, b=5, args=(7,8), kwargs={'a': 42}, extra_stuff='ignore'
    ... ) == {
    ... 'formula1': {'w': 1, 'x': 2, 'z': 3},
    ... 'mult': {'x': 2},
    ... 'addition': {'a': 4, 'b': 5},
    ... 'mysum':
    ...     {'kwargs': {'w': 1,
    ...                 'x': 2,
    ...                 'z': 3,
    ...                 'a': 42,
    ...                 'b': 5,
    ...                 'args': (7, 8),
    ...                 'extra_stuff': 'ignore'}}}

    """

    # FIXME: TODO: This does indeed change the signature, but not the functionality (position only still raise errors!)
    def normalize_func(self, func):
        return ch_func_to_all_pk(tuple_the_args(func))

    def __init__(self, *unnamed_funcs, **named_funcs):
        # The basic MultiFunc initialization
        super().__init__(*unnamed_funcs, **named_funcs)

        # The extra initialization for FlexFuncFanouts
        self.sigs = {name: Sig(func) for name, func in self.funcs.items()}
        self.normalized_funcs = {
            name: self.normalize_func(func) for name, func in self.funcs.items()
        }
        multi_func_sig = Sig.from_objs(*self.normalized_funcs.values())
        # TODO: Finish attempt to add **all_other_kwargs_ignored to the signature
        # multi_func_sig = (Sig.from_objs(
        #     *self.normalized_funcs.values(),
        #     Parameter(name='all_other_kwargs_ignored', kind=Parameter.VAR_KEYWORD)))
        multi_func_sig.wrap(self)
        # multi_func_sig.wrap(self.kwargs_for_func)

    def kwargs_for_func(self, *args, **kwargs):
        return dict(
            (name, self.sigs[name].source_kwargs(**kwargs))
            for name, func in self.funcs.items()
        )

    # TODO: Give it a signature (needs to be done in __init__)
    # TODO: Validation of inputs
    def __call__(self, *args, **kwargs):
        return dict(
            (name, self.sigs[name].source_kwargs(**kwargs))
            for name, func in self.funcs.items()
        )


class ParallelFuncs(MultiFunc):
    """Make a multi-channel function from a {name: func, ...} specification.

    >>> multi_func = ParallelFuncs(
    ...     say_hello=lambda x: f"hello {x}", say_goodbye=lambda x: f"goodbye {x}"
    ... )
    >>> multi_func({'say_hello': 'world', 'say_goodbye': 'Lenin'})
    {'say_hello': 'hello world', 'say_goodbye': 'goodbye Lenin'}

    :param spec: A map between a name (str) and a function associated to that name
    :return: A function that takes a dict as an (multi-channel) input and a dict as a
    (multi-channel) output

    Q: Why can I specify the specs both with named_funcs_dict and **named_funcs?
    A: Look at the ``dict(...)`` interface. You see the same thing there.
    Different reason though (here we assert that the keys don't overlap).
    Usually named_funcs is more convenient, but if you need to use keys that are not
    valid python variable names,
    you can always use named_funcs_dict to express that!

    >>> multi_func = ParallelFuncs({
    ...     'x+y': lambda d: f"sum is {d}",
    ...     'x*y': lambda d: f"prod is {d}"}
    ... )
    >>> multi_func({
    ...     'x+y': 5,
    ...     'x*y': 6
    ... })
    {'x+y': 'sum is 5', 'x*y': 'prod is 6'}

    You can also use both. Like with ``dict(...)``.

    Here's a more significant example.

    >>> chunkers = {
    ...     'a': lambda x: x[0] + x[1],
    ...     'b': lambda x: x[0] * x[1]
    ... }
    >>> featurizers = {
    ...     'a': lambda z: str(z),
    ...     'b': lambda z: [z] * 3
    ... }
    >>> multi_chunker = ParallelFuncs(**chunkers)
    >>> multi_chunker({'a': (1, 2), 'b': (3, 4)})
    {'a': 3, 'b': 12}
    >>> multi_featurizer = ParallelFuncs(**featurizers)
    >>> multi_featurizer({'a': 3, 'b': 12})
    {'a': '3', 'b': [12, 12, 12]}
    >>> my_pipe = Pipe(multi_chunker, multi_featurizer)
    >>> my_pipe({'a': (1, 2), 'b': (3, 4)})
    {'a': '3', 'b': [12, 12, 12]}

    #{'a': '(1, 2)', 'b': [(3, 4), (3, 4), (3, 4)]}

    """

    def _key_output_gen(self, d: dict):
        for key, func in self.funcs.items():
            yield key, func(d[key])

    def __call__(self, d: dict):
        return dict(self._key_output_gen(d))


# from collections.abc import ValuesView as BaseValuesView
# class MongoValuesView(ValuesView):
#     def __contains__(self, v):
#         m = self._mapping
#         cursor = m.mgc.find(filter=m._merge_with_filt(v), projection=())
#         return next(cursor, end_of_cursor) is not end_of_cursor
#
#     def __iter__(self):
#         m = self._mapping
#         return m.mgc.find(filter=m.filter, projection=m._getitem_projection)


class ContextFanout(MultiObj):
    """Encapsulates multiple objects into a single context manager that will enter and
    exit all objects that are context managers themselves.

    Context managers show up in situations where you need to have some setup and tear
    down before performing some tasks. It's what you get when you open a file to read
    or write in it, or open a data-base connection, etc.

    Sometimes you need to perform a task that involves more than one context managers,
    or even some objects that may or may not be context managers.
    What `ContextFanout` does for you is allow you to bundle all those (perhaps)
    context managers together, and use them as one single context manager.

    In python 3.10+ you can bundle contexts together by specifying a tuple of context
    managers, as such:

    ```python
    with (open('file.txt'), another_context_manager):
        ...
    ```

    But
    - Python will complain if one of the members of the tuple is not a context manager.
    - A tuple of context managers is not a context manager itself, it's just understood
    by the with (in python 3.10+).

    As an example, let's take two objects. One is a context manager, the other not.

    >>> from contextlib import contextmanager
    >>> @contextmanager
    ... def some_context_manager(x):
    ...     print('open')
    ...     yield f'x + 1 = {x + 1}'
    ...     print('close')
    ...
    >>> def not_a_context_manager(x):
    ...     return x - 1
    ...


    >>> c = ContextFanout(not_a_context_manager, some_context_manager(2))
    >>> list(c)
    ['not_a_context_manager', '_GeneratorContextManager']

    The name (chosen by `MultiObj.auto_namer`) '_GeneratorContextManager' isn't the best.
    Let's give an explicit name:

    >>> c = ContextFanout(not_a_context_manager, context=some_context_manager(2))
    >>> list(c.objects)
    ['not_a_context_manager', 'context']

    See from the prints that "with-ing" c triggers the enter and exit of 'context'

    >>> with c:
    ...     pass
    open
    close

    # Further, know that within the context's scope, a `ContextFanout`
    # instance will have the context managers it contains available, and having the
    # value it is supposed to have "under context".
    #
    # >>> with ContextFanout(not_a_context_manager, some_context_manager(2)) as (f, m):
    # ...     print(f(10))
    #
    #
    # >>> c = ContextFanout(not_a_context_manager, context=some_context_manager(2))
    #
    # >>> c = ContextFanout(
    # ...     some_context_manager=some_context_manager(2),
    # ...     not_a_context_manager=not_a_context_manager
    # ... )
    # >>> # first c doesn't have the some_context_manager attribute
    # >>> assert not hasattr(c, 'some_context_manager')
    # >>> with c:
    # ...     # inside the context, c indeed has the attribute, and it has the expected value
    # ...     assert c.some_context_manager == 'x + 1 = 3'
    # open
    # close
    # >>> # outside the context, c doesn't have the some_context_manager attribute any more again
    # >>> assert not hasattr(c, 'some_context_manager')
    #
    # If you don't specify a name for a given context manager, you'll still have access
    # to it via a hidden attribute ("_i" where i is the index of the object when
    # the `ContextFanout` instance was made.
    #
    # >>> c = ContextFanout(some_context_manager(10), not_a_context_manager)
    # >>> with c:
    # ...     assert c._0 == 'x + 1 = 11'
    # open
    # close

    """

    def __enter__(self):
        for name, obj in self.items():
            if hasattr(obj, '__enter__'):
                obj.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        for name, obj in self.items():
            if hasattr(obj, '__exit__'):
                obj.__exit__(exc_type, exc_val, exc_tb)



"""
Deprecated code I'm keeping around so I can at least reuse it's docs at some point.
"""

from functools import partial
from typing import Iterable, Callable, Any

from i2 import Sig, name_of_obj
from i2.signatures import extract_arguments


class Command:
    """A class that holds a `(caller, args, kwargs)` triple and allows one to execute
    `caller(*args, **kwargs)`

    :param func: A callable that will be called with (*args, **kwargs) argument.
    :param args: The positional arguments to call the func with.
    :param kwargs: The keyword arguments to call the func with.

    >>> c = Command(print, "hello", "world", sep=", ")
    >>> c()
    hello, world

    What happens (when a command is executed) if some of the arguments are commands
    themselves? Well, the sensible thing happens. These commands are executed.
    You can use this to define, declaratively, some pretty complex instructions, and
    only fetch the data you need and execute everything, once you're ready.

    >>> def show(a, b):
    ...     print(f"Showing this: {a=}, {b=}")
    >>> def take_five():
    ...     return 5
    >>> def double_val(val):
    ...     return val * 2
    >>> command = Command(
    ...     show,
    ...     Command(take_five),
    ...     b=Command(double_val, 'hello'),
    ... )
    >>> command
    Command(show, Command(take_five), b=Command(double_val, 'hello'))
    >>> command()
    Showing this: a=5, b='hellohello'

    Of course, as your use of Command gets more complex, you may want to subclass it
    and include some "validation" and "compilation" in the init.

    The usual way to call a function is to... erm... call it.
    But sometimes you want to do things differently.
    Like validate it, put it on a queue, etc.
    That's where specifying a different _caller will be useful.

    >>> class MyCommand(Command):
    ...     def _caller(self):
    ...         f, a, k = self.func, self.args, self.kwargs
    ...         print(f"Calling {f}(*{a}, **{k}) with result: {f(*a, **k)}")
    ...
    >>> c = MyCommand(print, "hello", "world", sep=", ")
    >>> c()
    hello, world
    Calling <built-in function print>(*('hello', 'world'), **{'sep': ', '}) with result: None

    """

    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    @classmethod
    def curried(cls, func, **kw_defaults):
        """Get an Command maker for a specific function, with defaults and signature!

        >>> def foo(x: str, y: int):
        ...     return x * y
        ...
        >>> foo('hi', 3)
        'hihihi'
        >>>
        >>> foo_command = Command.curried(foo, y=2)
        >>> Sig(foo_command)
        <Sig (x: str, y: int = 2)>
        >>> f = foo_command('hi', y=4)
        >>> f()
        'hihihihi'
        >>> ff = foo_command('hi')
        >>> ff
        Command(foo, 'hi')
        >>> ff()
        'hihi'

        """
        sig = Sig(func)
        sig = sig.ch_defaults(**kw_defaults)

        if kw_defaults:
            func = partial(func, **kw_defaults)

        curried_command_cls = partial(Command, func)
        return sig(curried_command_cls)

    def __repr__(self):
        def to_str(x, quote="'"):
            if isinstance(x, str):
                return quote + x + quote
            else:
                return str(x)

        args_str = ', '.join(to_str(a) for a in self.args)
        kwargs_str = ', '.join(f'{k}={to_str(v)}' for k, v in self.kwargs.items())
        if args_str and kwargs_str:
            sep = ', '
        else:
            sep = ''
        args_kwargs_str = args_str + sep + kwargs_str

        func_name = name_of_obj(self.func)
        if args_kwargs_str:
            return f'{type(self).__name__}({func_name}, {args_kwargs_str})'
        else:
            return f'{type(self).__name__}({func_name})'

    def _caller(self):
        return self.func(*self.args, **self.kwargs)

    def _args_with_executed_commands(self):
        for v in self.args:
            if isinstance(v, Command):
                v = v()  # if a command, execute it
            yield v

    def _kwargs_with_executed_commands(self):
        for k, v in self.kwargs.items():
            if isinstance(v, Command):
                v = v()  # if a command, execute it
            yield k, v

    def _caller(self):
        return self.func(
            *self._args_with_executed_commands(),
            **dict(self._kwargs_with_executed_commands()),
        )

    def __call__(self):
        return self._caller()


def extract_commands(
    funcs: Iterable[Callable],
    *,
    mk_command: Callable[[Callable, tuple, dict], Any] = Command,
    what_to_do_with_remainding='ignore',
    **kwargs,
):
    """

    :param funcs: An iterable of functions
    :param mk_command: The function to make a command object
    :param kwargs: The argname=argval items that the functions should draw from.
    :return:

    >>> def add(a, b: float = 0.0) -> float:
    ...     return a + b
    >>> def mult(x: float, y=1):
    ...     return x * y
    >>> def formula1(w, /, x: float, y=1, *, z: int = 1):
    ...     return ((w + x) * y) ** z
    >>> commands = extract_commands(
    ...     (add, mult, formula1), a=1, b=2, c=3, d=4, e=5, w=6, x=7
    ... )
    >>> for command in commands:
    ...     print(
    ...         f"Calling {command.func.__name__} with "
    ...         f"args={command.args} and kwargs={command.kwargs}"
    ...     )
    ...     print(command())
    ...
    Calling add with args=() and kwargs={'a': 1, 'b': 2}
    3
    Calling mult with args=() and kwargs={'x': 7}
    7
    Calling formula1 with args=(6,) and kwargs={'x': 7}
    13
    """
    extract = partial(
        extract_arguments,
        what_to_do_with_remainding=what_to_do_with_remainding,
        include_all_when_var_keywords_in_params=False,
        assert_no_missing_position_only_args=True,
    )

    if callable(funcs):
        funcs = [funcs]

    for func in funcs:
        func_args, func_kwargs = extract(func, **kwargs)
        yield mk_command(func, *func_args, **func_kwargs)


def commands_dict(
    funcs,
    *,
    mk_command: Callable[[Callable, tuple, dict], Any] = Command,
    what_to_do_with_remainding='ignore',
    **kwargs,
):
    """

    :param funcs:
    :param mk_command:
    :param kwargs:
    :return:

    >>> def add(a, b: float = 0.0) -> float:
    ...     return a + b
    >>> def mult(x: float, y=1):
    ...     return x * y
    >>> def formula1(w, /, x: float, y=1, *, z: int = 1):
    ...     return ((w + x) * y) ** z
    >>> d = commands_dict((add, mult, formula1), a=1, b=2, c=3, d=4, e=5, w=6, x=7)
    >>> d[add]()
    3
    >>> d[mult]()
    7
    >>> d[formula1]()
    13

    """
    if callable(funcs):
        funcs = [funcs]
    it = extract_commands(
        funcs,
        what_to_do_with_remainding=what_to_do_with_remainding,
        mk_command=mk_command,
        **kwargs,
    )
    return dict(zip(funcs, it))



"""Decorator tools"""
import inspect
from collections import defaultdict
from contextlib import suppress
from functools import wraps, partial
from inspect import Signature, signature, Parameter
from itertools import chain
from typing import Callable, Tuple, Dict, Any, TypeVar

from i2.signatures import Sig, kind_forgiving_func, name_of_obj

# keep the imports below here because might be referenced (or take care of refs)
# from i2.signatures import ch_signature_to_all_pk, params_of, copy_func

T = TypeVar('T')  # Can be anything


def identity(obj: T) -> T:
    return obj


# ---------------------------------------------------------------------------------------
# The following is an exact copy of the `functools.wraps`, but for some reason, it
# solves an issue of double_up_as_factory on jypyter.
# See description of problem here:
# https://stackoverflow.com/questions/73243479/mysterious-effect-a-decorator-that-breaks-juypyters-ability-to-tab-complete-a

from functools import WRAPPER_ASSIGNMENTS, WRAPPER_UPDATES, update_wrapper


def wraps(wrapped, assigned=WRAPPER_ASSIGNMENTS, updated=WRAPPER_UPDATES):
    return partial(update_wrapper, wrapped=wrapped, assigned=assigned, updated=updated)


# ---------------------------------------------------------------------------------------
class FuncFactory:
    """Make a function factory.

    but more convenient and helpful (e.g. is picklable, produces functions with
    signatures, etc.)

    One can use ``functools.partials`` to fix, or change, the defaults of
    arguments of a function ``func`` thereby creating a different function.

    >>> def foo(a, b, *, c=2) -> float:
    ...     return a * b + c
    >>> foo(10, 2)
    22
    >>> foo(10, b=2, c=3)
    23
    >>> from functools import partial
    >>> new_foo = partial(foo, b=2, c=3)  # change default c=3 and add one: b=2
    >>> new_foo(10)  # now the function can be called with one argument (couldn't before)
    23

    In essence, ``FuncFactory`` is equivalent to:

    ```
    FuncFactory = lambda func: lambda *args, **kwargs: partial(func, *args, **kwargs)
    ```

    but more convenient and helpful. For one, it doesn't use ``lambda``, so is picklable.
    It also has a more helpful signature:

    >>> factory = FuncFactory(foo)
    >>> factory
    <FuncFactory(foo)>(a, b, *, c=2) -> Callable[..., float]

    (Note that the repr even reuses ``foo``'s return annotation to tell us that our
    factory will return a callable that returns that type (if the annotation is a type).

    An instance of ``FuncFactory`` is a factory of functions, that is, it can make
    functions for you based on the instance's underlying ``func``:

    >>> f = factory(b=2, c=3)
    >>> f(10)
    23

    Recipe: Say you're normalizing some data accessor into callback functions and you
    want to create functions that provide a specific object when called (with no args).
    Sure, you can do this by specifying ``lambda: obj`` every time, but lambdas can be
    problematic (e.g. their not picklable).

    Here's another solution:

    >>> def identity(obj):
    ...     return obj
    >>> func_returning_obj = FuncFactory(identity)
    >>> get_42 = func_returning_obj(42)
    >>> get_42()
    42

    Note: A convenience property has been added to implement this recipe:

    >>> get_42, get_hello = map(FuncFactory.func_returning_obj, (42, 'hello'))
    >>> get_42()
    42
    >>> get_hello()
    'hello'

    """

    def __init__(self, func):
        self.func = func

        sig = Sig(func)
        factory_sig = Sig(sig, return_annotation=Callable[..., Any])
        # previous I did the following, but don't know why:
        # factory_sig = factory_sig - sig.names[0]
        if sig.return_annotation is not sig.empty:
            try:
                factory_sig = Sig(
                    factory_sig, return_annotation=Callable[..., sig.return_annotation]
                )
            except TypeError:
                pass

        self.__signature__ = factory_sig

    def __call__(self, *args, **kwargs):
        return partial(self.func, *args, **kwargs)

    def __repr__(self):
        return f'<FuncFactory({name_of_obj(self.func)})>{self.__signature__}'

    @classmethod
    def func_returning_obj(cls, obj):
        return cls(identity)(obj)


def _double_up_as_factory(wrapped=None, *args, __decorator_func=None, **kwargs):
    """Util for double_up_as_factory, ``__decorator_func`` to be partialized"""
    if args:
        raise RuntimeError(
            f'You need to specify decorator arguments as keyword-only.'
            f'You specified positional arguments: {args=}'
        )
    if wrapped is None:  # then we want a factory
        return partial(__decorator_func, **kwargs)
    else:
        return __decorator_func(wrapped, *args, **kwargs)


def double_up_as_factory(decorator_func):
    """Repurpose a decorator both as it's original form, and as a decorator factory.
    That is, from a decorator that is defined do ``wrapped_func = decorator(func, **params)``,
    make it also be able to do ``wrapped_func = decorator(**params)(func)``.

    Note: You'll only be able to do this if all but the first argument are keyword-only,
    and the first argument (the function to decorate) has a default of ``None`` (this is for your own good).
    This is validated before making the "double up as factory" decorator.

    >>> @double_up_as_factory
    ... def decorator(func=None, *, multiplier=2):
    ...     def _func(x):
    ...         return func(x) * multiplier
    ...     return _func
    ...
    >>> def foo(x):
    ...     return x + 1
    ...
    >>> foo(2)
    3
    >>> wrapped_foo = decorator(foo, multiplier=10)
    >>> wrapped_foo(2)
    30
    >>>
    >>> multiply_by_3 = decorator(multiplier=3)
    >>> wrapped_foo = multiply_by_3(foo)
    >>> wrapped_foo(2)
    9
    >>>
    >>> @decorator(multiplier=3)
    ... def foo(x):
    ...     return x + 1
    ...
    >>> foo(2)
    9

    Note that to be able to use double_up_as_factory, your first argument (the object to be wrapped) needs to default
    to None and be the only argument that is not keyword-only (i.e. all other arguments need to be keyword only).

    >>> @double_up_as_factory
    ... def decorator_2(func, *, multiplier=2):
    ...     '''Should not be able to be transformed with double_up_as_factory'''
    Traceback (most recent call last):
      ...
    AssertionError: First argument of the decorator function needs to default to None. Was <class 'inspect._empty'>
    >>> @double_up_as_factory
    ... def decorator_3(func=None, multiplier=2):
    ...     '''Should not be able to be transformed with double_up_as_factory'''
    Traceback (most recent call last):
      ...
    AssertionError: All arguments (besides the first) need to be keyword-only

    """

    def validate_decorator_func(decorator_func):
        first_param, *other_params = signature(decorator_func).parameters.values()
        assert first_param.default is None, (
            f'First argument of the decorator function needs to default to None. '
            f'Was {first_param.default}'
        )
        assert all(
            p.kind in {p.KEYWORD_ONLY, p.VAR_KEYWORD} for p in other_params
        ), f'All arguments (besides the first) need to be keyword-only'
        return True

    validate_decorator_func(decorator_func)

    return wraps(decorator_func)(
        partial(_double_up_as_factory, __decorator_func=decorator_func)
    )


# TODO: Review, doc, and make public.
#   Namely, it may be cleaner to have the `argname=argtrans, ...` interface
#   and offer utils to make an argtrans be the combination of a condition and a trans!
@double_up_as_factory
def _conditional_arg_trans(func=None, **condition_and_trans_of_argname):
    """

    See also: ensure_iterable_args, for a more complex example of how to use it.
    """

    @wraps(func)
    def _func(*args, **kwargs):
        sig = Sig(func)
        kwargs = sig.kwargs_from_args_and_kwargs(args, kwargs)
        for argname, (condition, trans) in condition_and_trans_of_argname.items():
            if argname in kwargs and condition(kwargs[argname]):
                # Note: Using tuple instead of list because quicker & safer (non-mutable)
                kwargs[argname] = trans(kwargs[argname])
        args, kwargs = sig.args_and_kwargs_from_kwargs(kwargs)
        return func(*args, **kwargs)

    return _func


def _tuplize(x):
    return (x,)


def _mk_type_condition(condition):
    """If condition is a type, make an isinstance function to check for that type"""
    curryable_isinstance = kind_forgiving_func(isinstance)
    if isinstance(condition, type) or isinstance(condition, tuple):
        type_condition = condition
        condition = partial(curryable_isinstance, class_or_tuple=type_condition)
    return condition


def ensure_iterable_args(func=None, **condition_of_argname):
    """Wrap a function so that specific arguments are assured to be iterable if
    they meet specific conditions.

    The condition, in the example below, is being a string.
    Note that in general, the condition needs to be a boolean function.
    The explicit form of our example would say `names=lambda x: isinstance(x, str)`,
    but `ensure_iterable_args` allows the convenience of just specifying the type,
    or a tuple of types, and the actually boolean function will be made for you.

    >>> @ensure_iterable_args(names=str)
    ... def greet_people(names, greeting='Hello'):
    ...     for name in names:
    ...         yield f"{greeting} {name}!"
    >>> assert list(greet_people(['Alice', 'Bob'])) == ['Hello Alice!', 'Hello Bob!']
    >>> assert list(greet_people('Alice')) == ['Hello Alice!']

    Note that to decorate a function, you can also use the form:

    >>> greet_people = ensure_iterable_args(greet_people, names=str)
    """
    condition_and_trans_of_argname = {
        argname: (_mk_type_condition(condition), _tuplize)
        for argname, condition in condition_of_argname.items()
    }
    return _conditional_arg_trans(func, **condition_and_trans_of_argname)


def transparently_wrapped(func):
    @wraps(func)
    def transparently_wrapped_func(*args, **kwargs):
        return func(args, **kwargs)

    return transparently_wrapped_func


def mk_args_kwargs_merger(func):
    """
    Make a function that will return a dict containing all {argname: argval} pairs from a function's call.
    That is, it merges all non-keyword arguments with the keyword-arguments, with the right name, so that
    the arguments can be handled more uniformly.
    :param func: The function that will be called, whose signature should be looked at to make the
        merging function
    :return: A function merge_args_and_kwargs(args, kwargs) that can be used to merge arguments

    >>> def func(a, b, c=3):
    ...     return a * (b + c)
    >>> merger = mk_args_kwargs_merger(func)
    >>> dict(merger([1], {'b': 10}))
    {'a': 1, 'b': 10}
    >>> dict(merger([], {'a': 1, 'b': 10}))
    {'a': 1, 'b': 10}
    >>> dict(merger([1, 10], {}))
    {'a': 1, 'b': 10}
    >>> dict(merger([], {}))
    {}
    >>> # Usage demo:
    >>> assert func(*[1], **{'b': 10}) == func(**merger([1], {'b': 10}))
    >>> assert func(*[], **{'a': 1, 'b': 10}) == func(**merger([], {'a': 1, 'b': 10}))
    >>> assert func(**{'a': 1, 'b': 10}) == func(**merger([], {'a': 1, 'b': 10}))
    """

    def merge_args_and_kwargs(args, kwargs):
        if len(args) > 0:
            return inspect.signature(func).bind_partial(*args, **kwargs).arguments
        else:
            return kwargs

    return merge_args_and_kwargs


def kwargs_for_func(*funcs, **kwargs):
    """
    :param funcs:
    :param kwargs:
    :return:

    >>> from i2.tests.objects_for_testing import formula1, sum_of_args, mult, add
    >>> def print_dict(d):  # just a util for this doctest
    ...     from pprint import pprint
    ...     pprint({k.__name__: d[k] for k in sorted(d, key=lambda x: x.__name__)})
    >>> print_dict(kwargs_for_func(formula1, mult, add,
    ...                           w=1, x=2, z=3, a=4, b=5)) # doctest: +NORMALIZE_WHITESPACE
    {'add': {'a': 4, 'b': 5},
     'formula1': {'w': 1, 'x': 2, 'z': 3},
     'mult': {'x': 2}}
    """
    return dict((func, Sig(func).source_kwargs(**kwargs)) for func in funcs)


def assert_attrs(attrs):
    """
    Asserts, at construction time, that the class contains a specific set of attributes
    :param attrs: An attribute name (string) or a list of attribute names whose existence needs to be enforced.
    :return: A class decorator that will enforce the existence of the attrs when an instance is made

    >>> @assert_attrs('foo')
    ... class A:
    ...     bar = 10
    ...
    >>> try:
    ...     a = A()
    ... except AttributeError:
    ...     print("AttributeError, as expected, because missing the foo attribute")
    AttributeError, as expected, because missing the foo attribute
    >>> @assert_attrs('foo')
    ... class B:
    ...     def foo(self): pass
    >>> b = B()
    >>>
    >>> class A:
    ...     bar = 10
    >>> class B:
    ...     def foo(self): pass
    >>>
    >>> @assert_attrs(['foo', 'bar'])
    ... class C(A, B):
    ...     pass
    >>> c = C()
    """
    if isinstance(attrs, str):
        attrs = [attrs]

    def _assert_attrs(klass):
        @wraps(klass)
        def get_instance(*args, **kw):
            for attr in attrs:
                if not hasattr(klass, attr):
                    raise AttributeError(
                        'class {} needs to have a {} attribute:'.format(
                            klass.__name__, attr
                        )
                    )
            return klass(*args, **kw)

        return get_instance

    return _assert_attrs


def preprocess_arguments(pre):
    """Apply a function to args, kwargs and use the transformed in the decorated function"""

    def decorator(func):
        def wrapper(*args, **kwargs):
            args, kwargs = pre(*args, **kwargs)
            return func(*args, **kwargs)

        return wraps(func)(wrapper)

    return decorator


def preprocess(pre):
    def decorator(func):
        if inspect.ismethod(func):

            def wrapper(self, *args, **kwargs):
                return func(self, pre(*args, **kwargs))

        else:

            def wrapper(*args, **kwargs):
                return func(pre(*args, **kwargs))

        return wraps(func)(wrapper)

    return decorator


def _return_annotation_of(func):
    """Return annotation of callable (if type, will return type systematically)

    >>> def foo() -> bool: ...
    >>> assert _return_annotation_of(foo) == bool
    >>> assert _return_annotation_of(zip) == zip
    >>> assert _return_annotation_of(print) == Parameter.empty
    """
    if isinstance(
        func, type
    ):  # TODO: Verify rule (are there commmon enough meta tricks that need to be handled?)
        return func
    else:
        try:
            return signature(func).return_annotation
        except ValueError:  # some builtins don't have signatures
            return Parameter.empty


class OutputPostProcessingError(RuntimeError):
    ...


def postprocess(post, caught_post_errors=(Exception,), verbose_error_message=False):
    """Add some post-processing after a function

    :param post: The function to apply to the output

    >>> list_range = postprocess(list)(range)
    >>> list_range(4)
    [0, 1, 2, 3]
    >>> sum_range = postprocess(sum)(range)
    >>> sum_range(4)
    6

    Note: The decorator also sticks the return annotation of the post function on the wrapped one.

    Use cases:

    - Changing a generator into a container returning function
        In many situations, writing a generator is simpler than writing a function
        that accumulates a list or a dict etc.
        So here, you just write the generator and tag this decorator on top, to get the same effect.

    >>> from inspect import signature
    >>> @postprocess(dict)
    ... def bar(x):
    ...     for i in range(x):
    ...         yield str(i), i
    >>> bar(3)
    {'0': 0, '1': 1, '2': 2}
    >>> signature(bar)
    <Signature (x) -> dict>
    >>>
    >>> @postprocess(list)
    ... def foo(x):
    ...     for i in range(x):
    ...         yield i
    >>> foo(3)
    [0, 1, 2]
    >>> from inspect import signature
    >>> signature(foo)
    <Signature (x) -> list>

    - Triggering something (like logging, or forwarding) when a function returns

    >>> def log_this(x):
    ...     print(f"Logging {x}")
    ...     return x
    >>> logged_foo = postprocess(log_this)(foo)
    >>> t = logged_foo(2)
    Logging [0, 1]
    >>> assert t == [0, 1]

    - Using a function that does a lot to make several functions that do less.
        (e.g. Extracting/making a python object from a function returning a raw http response_

    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            output = func(*args, **kwargs)
            try:
                return post(output)
            except caught_post_errors as e:
                msg = f'Error when postprocessing output with post func: {func}'
                if verbose_error_message:
                    msg += '\n' + f'  output={output}'
                    if (
                        isinstance(verbose_error_message, int)
                        and verbose_error_message > 1
                    ):
                        msg += (
                            '\n'
                            + '  which was obtained by func(*args, **kwargs) where:'
                        )
                        msg += (
                            '\n' + f'    args: {args}' + '\n' + f'    kwargs: {kwargs}'
                        )
                msg += '\n' + f'Error is: {e}'
                raise OutputPostProcessingError(msg)

        return_annot = _return_annotation_of(post)
        with suppress(
            ValueError
        ):  # intended to catch cases where wrapper doesn't have a signature
            wrapper_signature = signature(wrapper)
            sig = Signature(
                wrapper_signature.parameters.values(), return_annotation=return_annot,
            )
            wrapper.__signature__ = sig

        return wraps(func)(wrapper)

    return decorator


def input_output_decorator(preprocess=None, postprocess=None):
    """
    Makes a decorator that preprocesses inputs and postprocesses outputs.
    Use it if you want to transform the input of a function or method before calling it, or if you want
    to transform the returned value before returning it.
    :param preprocess: Function to be applied to input
    :param postprocess: Function to be applied to output
    :return: a decorator that preprocesses inputs and postprocesses outputs
    See also: preprocess and postprocess decorators if you need only to pre or post process!

    >>> # Examples with "normal functions"
    >>> def f(x=3):
    ...     '''Some doc...'''
    ...     return x + 10
    >>> ff = input_output_decorator()(f)
    >>> print((ff(5.0)))
    15.0
    >>> ff = input_output_decorator(preprocess=int)(f)
    >>> print((ff(5.0)))
    15
    >>> ff = input_output_decorator(preprocess=int, postprocess=lambda x: "Hello {}!".format(x))(f)
    >>> print((ff('5')))
    Hello 15!
    >>> ff = input_output_decorator(postprocess=lambda x: "Hello {}!".format(x))(f)
    >>> print((ff(5.0)))
    Hello 15.0!
    >>> print((ff.__doc__))
    Some doc...
    >>>
    >>> # examples with methods (bounded, class methods, static methods
    >>> class F:
    ...     '''This is not what you'd expect: The doc of the class, not the function'''
    ...     def __init__(self, y=10):
    ...         '''Initialize'''
    ...         self.y = y
    ...     def __call__(self, x=3):
    ...         '''Some doc...'''
    ...         return self.y + x
    ...     @staticmethod
    ...     def static_method(x, y):
    ...         return "What {} {} you have".format(x, y)
    ...     @classmethod
    ...     def class_method(cls, x):
    ...         return "{} likes {}".format(cls.__name__, x)
    >>>
    >>> f = F()
    >>> ff = input_output_decorator()(f)
    >>> print((ff(5.0)))
    15.0
    >>> ff = input_output_decorator(preprocess=int)(f)
    >>> print((ff(5.0)))
    15
    >>> ff = input_output_decorator(preprocess=int, postprocess=lambda x: "Hello {}!".format(x))(f)
    >>> print((ff('5')))
    Hello 15!
    >>> ff = input_output_decorator(postprocess=lambda x: "Hello {}!".format(x))(f)
    >>> print((ff(5.0)))
    Hello 15.0!
    >>> print((ff.__doc__))
    This is not what you'd expect: The doc of the class, not the function

    # >>>
    # >>> f.static_method = input_output_decorator(preprocess=lambda x: '"' + x + '"',
    # ...                                          postprocess=lambda x: x + '!!!')(f.static_method)
    # >>> print(ff.static_method('big', 'eyes'))
    # What big "eyes" you have!!!
    """

    def decorator(func):
        if preprocess and postprocess:

            def func_wrapper(*args, **kwargs):
                return postprocess(func(preprocess(*args, **kwargs)))

        elif preprocess:  # a preprocess but no postprocess

            def func_wrapper(*args, **kwargs):
                return func(preprocess(*args, **kwargs))

        elif postprocess:  # a postprocess but no preprocess

            def func_wrapper(*args, **kwargs):
                return postprocess(func(*args, **kwargs))

        else:  # neither pre nor post processing, so leave func as is
            return func

        return wraps(func)(func_wrapper)

    decorator.preprocess = preprocess
    decorator.postprocess = postprocess

    return decorator


def transform_args(dflt_trans_func=None, /, **trans_func_for_arg):
    """
    Make a decorator that transforms function arguments before calling the function.
    Works with plain functions and bounded methods.
    For example:
        * original argument: a relative path --> used argument: a full path
        * original argument: a pickle filepath --> used argument: the loaded object
    :param rootdir: rootdir to be used for all name arguments of target function
    :param name_arg: the position (int) or argument name of the argument containing the name
    :return: a decorator

    >>> # Example with a plain function
    >>> def f(a, b, c='default_c'):
    ...     return "a={a}, b={b}, c={c}".format(a=a, b=b, c=c)
    >>> def prepend_root(x):
    ...     return 'ROOT/' + x
    >>>
    >>> def test(f):
    ...     assert f('foo', 'bar', 3) == 'a=foo, b=bar, c=3'
    ...     ff = transform_args()(f)  # no transformation specification, so function is unchanged
    ...     assert ff('foo', 'bar', c=3) == 'a=foo, b=bar, c=3'
    ...     ff = transform_args(a=prepend_root)(f)  # prepend root to a
    ...     assert ff('foo', c=3, b='bar') == 'a=ROOT/foo, b=bar, c=3'  # note: testing different order of args
    ...     ff = transform_args(b=prepend_root)(f)  # prepend root to b
    ...     assert ff(c=3, b='bar', a='foo') == 'a=foo, b=ROOT/bar, c=3'  # note: testing different order of args
    ...     ff = transform_args(a=prepend_root, b=prepend_root)(f)  # prepend root to a and b
    ...     assert ff('foo', 'bar', 3) == 'a=ROOT/foo, b=ROOT/bar, c=3'
    ...     assert ff('foo', 'bar') == 'a=ROOT/foo, b=ROOT/bar, c=default_c'  # defaults still work
    >>>
    >>> test(f)
    >>>
    >>> # Example with bounded method, wrapping from instance
    >>> class A:
    ...     def __init__(self, sep=''):
    ...         self.sep = sep
    ...     def f(self, a, b, c='default_c'):
    ...         return f"a={a}{self.sep} b={b}{self.sep} c={c}"
    >>>
    >>> a = A(sep=',')
    >>> test(a.f)
    >>>
    >>> # Example with bounded method, wrapping from class
    >>> A.f = transform_args(a=prepend_root, b=prepend_root)(A.f)
    >>> a = A(sep=',')
    >>> assert a.f('foo', 'bar', 3) == 'a=ROOT/foo, b=ROOT/bar, c=3'
    >>> assert a.f('foo', 'bar') == 'a=ROOT/foo, b=ROOT/bar, c=default_c'  # defaults still work
    """

    def transform_args_decorator(func):
        get_kwargs = mk_args_kwargs_merger(func)

        if (
            len(trans_func_for_arg) == 0 and not dflt_trans_func
        ):  # if no transformations were specified...
            return func  # just return the function itself
        elif dflt_trans_func is not None:
            assert callable(
                dflt_trans_func
            ), 'The dflt_trans_func needs to be a callable'

            @wraps(func)
            def transform_args_wrapper(*args, **kwargs):
                val_of_argname = get_kwargs(args, kwargs)
                val_of_argname = {
                    argname: dflt_trans_func(val)
                    for argname, val in val_of_argname.items()
                }

                # apply transform functions to argument values
                return func(**val_of_argname)

            return transform_args_wrapper
        else:

            @wraps(func)
            def transform_args_wrapper(*args, **kwargs):
                # get a {argname: argval, ...} dict from *args and **kwargs
                # Note: Didn't really need an if/else here but I am assuming that...
                # Note: ... getcallargs gives us an overhead that can be avoided if there's only keyword args.

                val_of_argname = get_kwargs(args, kwargs)

                for argname, trans_func in trans_func_for_arg.items():
                    if argname in val_of_argname:
                        val_of_argname[argname] = trans_func(val_of_argname[argname])
                # apply transform functions to argument values
                return func(**val_of_argname)

            return transform_args_wrapper

    transform_args_decorator.dflt_trans_func = dflt_trans_func
    transform_args_decorator.trans_func_for_arg = trans_func_for_arg

    return transform_args_decorator


def wrap_method_output(wrapper_func):
    def _wrap_output(wrapped):
        @wraps(wrapped)
        def _wrapped(self, *args, **kwargs):
            return wrapper_func(wrapped(self, *args, **kwargs))

        return _wrapped

    return _wrap_output


def wrap_class_methods(
    _return_a_copy_of_the_class=True,
    _raise_error_if_non_existent_method=True,
    **wrapper_for_method,
):
    """
    Make a decorator that wraps specific methods.

    IMPORTANT: The decorator will by default return a copy of the class. This might incur some run time overhead.
    If this is desirable, for example, when you want to create several decorations of a same class.
    If you want to change the class itself (e.g. you're only loading it once in a module, and decorating it), then
    specify _return_a_copy_of_the_class=False

    Note that _return_a_copy_of_the_class=True has a side effect of building russian dolls of essentially subclasses
    of the class, which may have some undesirable results if repeated too many times.

    :param _return_a_copy_of_the_class: Specifies whether to
        return a copy of the class (_return_a_copy_of_the_class=True, the default),
        or change the actual loaded class itself (_return_a_copy_of_the_class=False)
    :param wrapper_for_method: method_name=wrapper_function pairs.
    :return: A class wrapper. That is, a decorator that takes a class and returns a decorated version of it
        (or decaorates "in-place" if _return_a_copy_of_the_class=False

    SEE ALSO:
        * wrap_method_output: The function that is called for every method we wrap.
        * transform_class_method_input_and_output: A wrap_class_methods that is specialized for input arg and output
            transformation.

    >>> from functools import wraps
    >>> class A:
    ...     def __init__(self, a=10):
    ...         self.a = a
    ...     def add(self, x):
    ...         return self.a + x
    ...     def multiply(self, x):
    ...         return self.a * x
    ...
    >>> a = A()
    >>> a.add(2)
    12
    >>> a.multiply(2)
    20
    >>>
    >>> def log_calls(func):
    ...     name = func.__name__
    ...     @wraps(func)
    ...     def _func(self, *args, **kwargs):
    ...         print("Calling {} with\\n  args={}\\n  kwargs={}".format(name, args, kwargs))
    ...         return func(self, *args, **kwargs)
    ...     return _func
    ...
    >>> AA = wrap_class_methods(**{k: log_calls for k in ['add', 'multiply']})(A)
    >>> a = AA()
    >>> a.add(x=3)
    Calling add with
      args=()
      kwargs={'x': 3}
    13
    >>> a.multiply(3)
    Calling multiply with
      args=(3,)
      kwargs={}
    30
    """

    def class_wrapper(cls):
        if _return_a_copy_of_the_class:
            _cls = type('_' + cls.__name__, cls.__bases__, dict(cls.__dict__))
            # class _cls(cls):
            #     pass
        else:
            _cls = cls
        for method, wrapper in wrapper_for_method.items():
            if hasattr(_cls, method):
                setattr(_cls, method, wrapper(getattr(_cls, method)))
            elif _raise_error_if_non_existent_method:
                raise ValueError(
                    f"{getattr(_cls, '__name__', str(cls))} has no '{method}' method!"
                )
        return _cls

    return class_wrapper


def mk_input_and_output_method_wrapper(method_output_trans=None, **arg_trans):
    def wrap_method(method_func):
        wrapped_method = transform_args(**arg_trans)(method_func)
        if method_output_trans is not None:
            return wrap_method_output(method_output_trans)(wrapped_method)
        else:
            return wrapped_method

    return wrap_method


def transform_class_method_input_and_output(
    cls, method, method_output_trans=None, **arg_trans
):
    wrapped_method = transform_args(**arg_trans)(getattr(cls, method))
    if method_output_trans is not None:
        setattr(
            cls, method, wrap_method_output(method_output_trans)(wrapped_method),
        )
    else:
        setattr(cls, method, wrapped_method)


def wrap_class_methods_input_and_output(
    _return_a_copy_of_the_class=True,
    _raise_error_if_non_existent_method=True,
    **method_trans_spec,
):
    """
    Make a decorator that wraps specific methods, transforming specific argument values a nd output values.

    IMPORTANT: The decorator will by default return a copy of the class. This might incur some run time overhead.
    If this is desirable, for example, when you want to create several decorations of a same class.
    If you want to change the class itself (e.g. you're only loading it once in a module, and decorating it), then
    specify _return_a_copy_of_the_class=False

    :param _return_a_copy_of_the_class: Specifies whether to
        return a copy of the class (_return_a_copy_of_the_class=True, the default),
        or change the actual loaded class itself (_return_a_copy_of_the_class=False)
    :param method_trans_spec: method_name=trans_specs_for_method pairs.
        The trans_specs_for_method is a dict that is understood by transform_class_method_input_and_output.
        Except for one special case, it's keys are argument names and values are callables to call on those
        arguments' values.
        The special case is method_output_trans. This specifies that the callable it points to should be called
        on output of method. Here's one recipe for outputs: If the output of a function is an iterable and you want
        to apply a function trans to each element of the output, specify method_output_trans=lambda x: map(trans, x).
    :return: A wrapped class

    SEE ALSO:
        * mk_method_trans_spec_from_methods_specs_dict: a utility to make method_trans_spec more easily
        * transform_class_method_input_and_output: The function that is called for every method we wrap.

    In the following, we will show two examples.
    - The first is a toy example to demonstrate the basic functionality.
    - The second demonstrates a more involved case, but is still a silly example.
    - The third demonstrates more the type of application we'd use wrap_class_methods_input_and_output for in real life.

    # FIRST EXAMPLE
    We make an Ops class that wraps Counter, allowing one to add items and show the counts of items added.

    >>> from collections import UserDict
    >>> import re
    >>> from collections import Counter
    >>>
    >>> class Ops:
    ...     def __init__(self):
    ...         self.counter = Counter()
    ...     def add_item(self, item):
    ...         self.counter.update({item: 1})
    ...     def show(self):
    ...         return self.counter
    >>> # Here's an example of what Ops does
    >>> ops = Ops()
    >>> for item in ['this', 'is', 'that', 'and', 'that', 'is', 'this']:
    ...     ops.add_item(item)
    ...
    >>> ops.show()
    Counter({'this': 2, 'is': 2, 'that': 2, 'and': 1})
    >>>
    >>> # But say we don't want to count actual words added, but just the first two letters of these words,
    >>> # and say we want to show() to return the dict, not the Counter.
    >>> NewOps = wrap_class_methods_input_and_output(
    ...     _return_a_copy_of_the_class=False,
    ...     add_item=dict(item=lambda x: x[:2]),  # intercept items fed to add_item and keep only 2 first letters
    ...     show=dict(method_output_trans=dict)  # intercept output of show method, converting to dict
    ... )(Ops)
    >>> # let's try it out!
    >>> ops = NewOps()
    >>> for item in ['this', 'is', 'that', 'and', 'that', 'is', 'this']:
    ...     ops.add_item(item)
    ...
    >>> ops.show()
    {'th': 4, 'is': 2, 'an': 1}
    >>> # See that we specified _return_a_copy_of_the_class=False?
    >>> # Now look at what happens if we try to use Ops, the original class, again. It behaves like NewOps.
    >>> # That's usually not the behavior we want, so be careful!
    >>> ops = Ops()
    >>> for item in ['this', 'is', 'that', 'and', 'that', 'is', 'this']:
    ...     ops.add_item(item)
    ...
    >>> ops.show()
    {'th': 4, 'is': 2, 'an': 1}
    >>>
    >>>

    # SECOND EXAMPLE

    Wrap a dict (or rather, the safer collections.UserDict), doing weird things to the input and output
    keys and values

    >>> val_in_trans = lambda x: 'hello {}'.format(x)  # prepend "hello " to incoming values
    >>> val_out_trans = lambda x: re.sub('hello', 'hi', x)  # replace "hello" by "hi" in output values
    >>> key_in_trans = lambda x: '__' + x  # prepend incoming keys with double underscore
    >>> key_out_trans = lambda x: x[2:]  # remove the first two characters (underscores) from keys when output
    >>>
    >>> methods_specs_dict = {
    ...     ('__contains__', '__getitem__', '__setitem__', '__delitem__'): dict(key=key_in_trans),
    ...     '__setitem__': dict(item=val_in_trans),
    ...     '__iter__': dict(method_output_trans=lambda x: map(key_out_trans, x)),
    ...     '__getitem__': dict(method_output_trans=val_out_trans)
    ... }
    >>>
    >>> methods_specs_dict = mk_method_trans_spec_from_methods_specs_dict(methods_specs_dict)
    >>>
    >>> @wrap_class_methods_input_and_output(**methods_specs_dict)
    ... class AA(UserDict):
    ...     pass
    ...
    >>> aa = AA()
    >>> aa['foo'] = 'shoo'  # store 'shoo' under 'foo'
    >>> # the __str__ method isn't wrapped, so we see the actual STORED keys and values
    >>> # we see that __foo, not foo is the actual key, and "hello shoo" the value:
    >>> assert str(aa) == "{'__foo': 'hello shoo'}"
    >>> assert 'foo' in aa  # yet from the interface, it looks like 'foo' is a key of aa...
    >>> assert '__foo' not in aa  # ... and '__foo' is not a key.
    >>> aa['foo'] = 'bar'  # let's replace the value of 'foo'
    >>> assert str(aa) == "{'__foo': 'hello bar'}"  # see what's stored
    >>> aa['star'] = 'wars'  # let's add another
    >>> assert list(aa) == ['foo', 'star']  # what are the keys? (this uses __iter__ under the hood)
    >>> # In the following, we'll use methods keys(), values(), and items(), none of which we wrapped.
    >>> # And yet, they work as expected, since they pass on their work to methods we wrapped.
    >>> assert list(aa.keys()) == ['foo', 'star']  # another way to get keys
    >>> # see here that when we ask for values, we don't get what we asked to store, ...
    >>> # ... nor what is actually stored, but something else
    >>> assert list(aa.values()) == ['hi bar', 'hi wars']
    >>> assert str(list(aa.items())) == "[('foo', 'hi bar'), ('star', 'hi wars')]"  # the keys and values we get from items()
    >>> assert str(aa) == "{'__foo': 'hello bar', '__star': 'hello wars'}"  # what is actually stored
    >>> del aa['foo']  # testing deletion of a key
    >>> assert str(aa) == "{'__star': 'hello wars'}"  # it worked!
    >>>
    >>>

    # THIRD EXAMPLE

    Here again, we'll wrap UserDict. But instead of being silly, we'll pretend we need to store waveforms
    in binary format (so input values will have to be wrapped), but still retrieving these waveforms as lists
    (so output values will have to be wrapped).
    Additionally, we'll pretend we're working with wav files within some root directory, but don't
    want the root dir or the '.wav' extension to appear in our keys. So we'll have to wrap input and output keys.
    Of course, this is just pretend. Don't use this with real waveforms. It won't work.

    >>> root = '/ROOT/DIR/'
    >>> abs_path_of_rel_path = lambda rel_path: root + rel_path + '.wav'  # transform a relative path to an absolute one
    >>> rel_path_of_abs_path = lambda x: x.replace(root, '').replace('.wav', '')  # transform an absolute path to a relative one
    >>> list_to_bytes = bytes
    >>> bytes_to_list = list
    >>>
    >>> methods_specs_dict = {
    ...     ('__contains__', '__getitem__', '__setitem__', '__delitem__'): dict(key=abs_path_of_rel_path),
    ...     '__setitem__': dict(item=list_to_bytes),
    ...     '__iter__': dict(method_output_trans=lambda x: map(rel_path_of_abs_path, x)),
    ...     '__getitem__': dict(method_output_trans=bytes_to_list)
    ... }
    >>>
    >>> methods_specs_dict = mk_method_trans_spec_from_methods_specs_dict(methods_specs_dict)
    >>>
    >>> @wrap_class_methods_input_and_output(**methods_specs_dict)
    ... class Wf(UserDict):
    ...     pass
    ...
    >>> year = [2, 0, 1, 9]
    >>> down = [5, 4, 3, 2, 1]
    >>>
    >>> wf = Wf()
    >>> wf['year'] = year
    >>> print(str(wf).replace("b'", "'"))
    {'/ROOT/DIR/year.wav': '\\x02\\x00\\x01\\t'}
    >>> 'year' in wf
    True
    >>> wf['down'] = down
    >>> print(str(wf).replace("b'", "'"))
    {'/ROOT/DIR/year.wav': '\\x02\\x00\\x01\\t', '/ROOT/DIR/down.wav': '\\x05\\x04\\x03\\x02\\x01'}
    >>> list(wf.keys())
    ['year', 'down']
    >>> list(wf.values())
    [[2, 0, 1, 9], [5, 4, 3, 2, 1]]
    >>> list(wf.items())
    [('year', [2, 0, 1, 9]), ('down', [5, 4, 3, 2, 1])]
    >>> len(wf)
    2
    >>> del wf['year']
    >>> len(wf)
    1
    >>> list(wf.items())
    [('down', [5, 4, 3, 2, 1])]
    """

    wrapper_for_method = {
        method: mk_input_and_output_method_wrapper(**method_trans)
        for method, method_trans in method_trans_spec.items()
    }
    return wrap_class_methods(
        _return_a_copy_of_the_class=_return_a_copy_of_the_class,
        _raise_error_if_non_existent_method=True,
        **wrapper_for_method,
    )

    # def class_wrapper(cls):
    #     if _return_a_copy_of_the_class:
    #         class _cls(cls):
    #             pass
    #     else:
    #         _cls = cls
    #     for method, method_trans in method_trans_spec.items():
    #         if hasattr(_cls, method):
    #             transform_class_method_input_and_output(_cls, method, **method_trans)
    #         elif _raise_error_if_non_existent_method:
    #             if hasattr(cls, '__name__'):
    #                 class_name = cls.__name__
    #             else:
    #                 class_name = str(cls)
    #             raise ValueError("{} has no '{}' method!".format(class_name, method))
    #     return _cls
    #
    # return class_wrapper


def add_method(obj, method_func, method_name=None, class_name=None):
    """
    Dynamically add a method to an object.

    :param obj: The object to add a method to
    :param method_func: The function to use as a method. The first argument must be the object itself
        (usually called self)
    :param method_name: The desired function name. If None, will take method_func.__name__
    :param class_name:  The desired class name. If None, will take type(obj).__name__
    :return: the object, but with the additional method (or a different function for it)

    >>> class A:
    ...     def __init__(self, x=10):
    ...         self.x = x
    >>> def times(self, y):
    ...     return self.x * y
    >>> def plus(self, y):
    ...     return self.x + y
    >>> a = A(x=10)
    >>> a = add_method(a, plus, '__call__')  # add a __call__ method, assigning it to plus
    >>> a(2)
    12
    >>> a = add_method(a, times, '__call__')  # reassign the __call__ method to times instead
    >>> a(2)
    20
    >>> a = add_method(a, plus, '__getitem__')  # assign the method __getitem__ to plus
    >>> a[2]  # see that it works
    12
    >>> a(2)  # and that we still have our __call__ method
    20
    """
    if isinstance(method_func, str):
        method_name = method_func
        method_func = getattr(obj, method_name)
    if method_name is None:
        method_name = method_func.__name__

    base = type(obj)

    if class_name is None:
        class_name = base.__name__
    bases = (base.__bases__[1:]) + (base,)
    bases_names = set(map(lambda x: x.__name__, bases))
    if class_name in bases_names:
        for i in range(6):
            class_name += '_'
            if not class_name in bases_names:
                break
        else:
            raise ValueError(
                "can't find a name for class that is not taken by bases. Consider using explicit name"
            )

    new_keys = set(dir(obj)) - set(chain(*[dir(b) for b in bases]))

    d = {a: getattr(obj, a) for a in new_keys}
    d[method_name] = method_func

    return type(class_name, bases, d)()


def transform_instance_method_input_and_output(
    obj, method, method_output_trans=None, **arg_trans
):
    from warnings import warn

    warn('Not sure transform_instance_method_input_and_output works yet')
    wrapped_method = transform_args(**arg_trans)(getattr(type(obj), method))
    if method_output_trans is not None:
        obj = add_method(
            obj,
            wrap_method_output(method_output_trans)(wrapped_method),
            method_name=method,
        )
    else:
        obj = add_method(obj, wrapped_method, method_name=method)
    return obj


def wrap_instance_methods(
    _return_a_copy_of_the_class=True,
    _raise_error_if_non_existent_method=True,
    **method_trans_spec,
):
    def obj_wrapper(obj):
        for method, method_trans in method_trans_spec.items():
            if hasattr(obj, method):
                obj = transform_instance_method_input_and_output(
                    obj, method, **method_trans
                )
            elif _raise_error_if_non_existent_method:
                if hasattr(obj.__class__, '__name__'):
                    class_name = obj.__name__
                else:
                    class_name = str(obj)
                raise ValueError("{} has no '{}' method!".format(class_name, method))
        return obj

    return obj_wrapper


def mk_method_trans_spec_from_methods_specs_dict(methods_specs_dict):
    """
    Utility to make inputs for wrap_class_methods_input_and_output more easily.
    :param methods_specs_dict: a dict where
        keys are method names (either a single string, or a tuple of strings)
        values are the trans_spec dicts that should be associated to those methods
    :return: A dict in the method_trans_spec (input of wrap_class_method) format.

    >>> methods_specs_dict = {}
    >>> methods_specs_dict['foo'] = {'x': str, 'y': int}
    >>> methods_specs_dict[('foo', 'bar')] = {'z': list, 'method_output_trans': float}
    >>> methods_specs_dict[('bar', )] = {'zz': int}
    >>> method_trans_spec = mk_method_trans_spec_from_methods_specs_dict(methods_specs_dict)
    >>> list(method_trans_spec.keys())
    ['foo', 'bar']
    >>> method_trans_spec['foo']
    {'x': <class 'str'>, 'y': <class 'int'>, 'z': <class 'list'>, 'method_output_trans': <class 'float'>}
    >>> method_trans_spec['bar']
    {'z': <class 'list'>, 'method_output_trans': <class 'float'>, 'zz': <class 'int'>}
    """
    method_trans_spec = defaultdict(dict)
    for methods, specs in methods_specs_dict.items():
        if isinstance(methods, str):
            methods = (methods,)
        for method in methods:
            method_trans_spec[method].update(specs)
    return dict(method_trans_spec)


Args = Tuple
Kwargs = Dict
WhatToLog = Callable[[Callable, Args, Kwargs], Any]


def _special_str(x: Any, max_len=100) -> str:
    """A util function for _call_signature"""
    if isinstance(x, str):
        return "'" + x + "'"
    else:
        x_str = str(x)
        if len(x_str) > max_len:
            type_str = getattr(type(x), '__name__', str(type(x)))
            if hasattr(x, '__repr__'):
                value_str = x.__repr__()
            else:
                value_str = x_str
            x_str = '{}({}...)'.format(type_str, value_str[:20])
        return x_str


def _call_signature(func: Callable, args: Args, kwargs: Kwargs) -> str:
    """
    A util to make a string representation of a call of a function func with given args and kwargs.
    Meant to be the default mk_log_str of mk_call_logger.
    :param func: A callable
    :param args: A tuple of positional arguments
    :param kwargs: A dict of key=val arguments
    :return: A string to represent all of that.

    >>> args = (2, 'sdf', list(range(1000)))
    >>> kwargs = {'z': 'boo', 'zzz': 10}
    >>> print(_call_signature(_call_signature, args, kwargs))
    _call_signature(2, 'sdf', list([0, 1, 2, 3, 4, 5, 6...), z='boo', zzz=10)
    """
    args_signature = ', '.join(map(_special_str, args))
    kwargs_signature = ', '.join(
        ('{}={}'.format(k, _special_str(v)) for k, v in kwargs.items())
    )
    return '{func_name}({signature})'.format(
        func_name=func.__name__,
        signature=', '.join([args_signature, kwargs_signature]),
    )


def mk_call_logger(
    logger=print,
    what_to_log: WhatToLog = _call_signature,
    log_output=False,
    func_is_bounded=False,
):
    """
    Makes a decorator that logs each call to the wrapped function.
    :param logger: The actual function that logs stuff. Default is print. The "stuff" it logs is given by
        the what_to_log argument (a function).
    :param what_to_log: A function taking inputs (func, args, kwargs) of the call, and returning something to log
        (usually, and by default, a string)
    :param func_is_bounded: Whether the function is bounded (like a method) or not
    :return: A decorator

    >>> # Example of use on (unbounded) function, with default args
    >>> @mk_call_logger()
    ... def useless_computation(x, y=2, z='foo'):
    ...     return z * (x + y)
    ...
    >>> _ = useless_computation(3, y=1, z='ha')
    useless_computation(3, y=1, z='ha')

    The same example, but with output logging too

    >>> @mk_call_logger(log_output=True)
    ... def useless_computation(x, y=2, z='foo'):
    ...     return z * (x + y)
    >>> _ = useless_computation(3, y=1, z='ha')
    useless_computation(3, y=1, z='ha')
    -> hahahaha

    And now a bit more involved...

    >>>
    >>> # Example of use on class method, with a different what_to_log function.
    >>> class A:
    ...     def __init__(self, a=10):
    ...         self.a = a
    ...     def add(self, x):
    ...         return self.a + x
    ...     def multiply(self, x):
    ...         return self.a * x
    ...
    >>> def _name_args_kwargs(func, args, kwargs) -> str:
    ...     return "Calling {} with\\n  args={}\\n  kwargs={}".format(func.__name__, args, kwargs)
    ...
    >>>
    >>> log_calls = mk_call_logger(what_to_log=_name_args_kwargs, func_is_bounded=True)
    >>> for method in ['add', 'multiply']:
    ...     A_method = getattr(A, method)
    ...     setattr(A, method, mk_call_logger(what_to_log=_name_args_kwargs, func_is_bounded=True)(A_method))
    ...
    >>>
    >>> a = A()
    >>> a.add(x=2)
    Calling add with
      args=()
      kwargs={'x': 2}
    12
    >>> a.multiply(2)
    Calling multiply with
      args=(2,)
      kwargs={}
    20
    """
    if log_output is True:
        log_output = '-> {}'.format
    assert log_output is False or callable(log_output)

    if not func_is_bounded:

        def log_calls(func):
            @wraps(func)
            def _func(*args, **kwargs):
                logger(what_to_log(func, args, kwargs))
                out = func(*args, **kwargs)
                log_output and logger(log_output(out))
                return out

            _func._logged_with = logger

            return _func

    else:

        def log_calls(func):
            @wraps(func)
            def _func(self, *args, **kwargs):
                logger(what_to_log(func, args, kwargs))
                out = func(self, *args, **kwargs)
                log_output and logger(log_output(out))
                return out

            _func._logged_with = logger

            return _func

    return log_calls


def get_callable_from_factory_if_no_arguments(func_or_factory_thereof: Callable):
    """Will return the input itself if it's a callable with at least one argument.
    If not, it will consider it to be a factory, call it to get the actual
    callable object that the user presumably is seeking to get"""
    assert callable(
        func_or_factory_thereof
    ), f'{func_or_factory_thereof} needs to be callable'
    if len(Sig(func_or_factory_thereof)) == 0:
        # if func_or_factory_thereof has no arguments, assume it's a factory
        func = func_or_factory_thereof()
        # and make sure that now we have arguments
        if not isinstance(func, Callable) or not len(Sig(func)) > 0:
            raise ValueError(
                'Your func_or_factory_thereof had no arguments, so I assumed it '
                "was a factory, called it, but that didn't produce a "
                "callable with at least one argument. So I don't know what to do."
            )
    else:
        func = func_or_factory_thereof
    return func


## This one didn't actually handle position only correctly (just signature)
# def old_ch_func_to_all_pk(func):
#     """Returns a copy of the function where all arguments are of the PK kind.
#     (PK: Positional_or_keyword)
#
#     :param func: A callable
#     :return:
#
#     >>> from py2http.decorators import signature, ch_func_to_all_pk
#     >>>
#     >>> def f(a, /, b, *, c=None, **kwargs): ...
#     ...
#     >>> print(signature(f))
#     (a, /, b, *, c=None, **kwargs)
#     >>> ff = old_ch_func_to_all_pk(f)
#     >>> print(signature(ff))
#     (a, b, c=None, **kwargs)
#     >>> def g(x, y=1, *args, **kwargs): ...
#     ...
#     >>> print(signature(g))
#     (x, y=1, *args, **kwargs)
#     >>> gg = old_ch_func_to_all_pk(g)
#     >>> print(signature(gg))
#     (x, y=1, args=(), **kwargs)
#     """
#     func = tuple_the_args(func)
#     sig = signature(func)
#     func.__signature__ = ch_signature_to_all_pk(sig)
#     return func



"""Error objects"""
import dataclasses
from typing import Callable, Mapping, Union, Any, Type
from functools import partial
from contextlib import AbstractContextManager


class DataError(Exception):
    pass


class DuplicateRecordError(DataError):
    pass


class NotFoundError(DataError):
    pass


class AuthorizationError(Exception):
    pass


class ForbiddenError(AuthorizationError):
    pass


class InputError(Exception):
    pass


class ModuleNotFoundIgnore:
    """Context manager meant to ignore import errors.
    The use case in mind is when we want to condition some code on the existence of some package.
    """

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is ModuleNotFoundError:
            pass
        return True


OnErrorCallback = Union[None, str, Callable[[], Any]]
ExceptionType = Type[BaseException]


def log_and_return(msg, logger=print):
    logger(msg)
    return msg


class InterruptWithBlock(BaseException):
    """To be used to interrupt the march of a with"""


# Note: Can be extended to have more precise handled conditions and callbacks
#  (involving exc_val and exc_tb)
# Note: Efforts towards a more general version here:
#   https://github.com/thorwhalen/ut/blob/33c20ce76fe0f9dcc6aa197b0a2dbbf3d7b1d5be/errors.py#L90
@dataclasses.dataclass
class HandleExceptions(AbstractContextManager):
    """A context manager that catches and (specifically) handles specific exceptions.

    It takes one argument: A dict (or mapping) of exception type keys and callback
    values. If within a with block, the particular (listed) exception happens,
    the callback is called and it's returned value is assigned to the
    `HandleExceptions` instance's `.exit_value` attribute.
    That attribute will only exist if the with block existed with an exception
    caught by `HandleExceptions`.

    A callback is an argument-less function. If you need to specify arguments, you can
    envoke the command pattern, using `functools.partial` to make a argument-less
    function. See in the example below how we ask `HandleExceptions` to print a
    specific string if a `ZeroDivisionError` happens:

    >>> from functools import partial
    >>> def print_and_return(msg):
    ...     print(msg)
    ...     return msg
    >>> with HandleExceptions({
    ...     ZeroDivisionError: partial(print_and_return, "You interrupted me"),
    ...     KeyboardInterrupt: lambda: 'imagine this is code to notify someone'
    ... }) as he:
    ...     print('This works')
    ...     0 / 0
    ...
    This works
    You interrupted me

    You can check if the context exited with a handled exception, and if so, what
    the callback returned value was.

    >>> he.exited_with_handled_exception()
    True
    >>> he.exit_value
    'You interrupted me'

    Also available, whether the exception was a handled one or not: The exception
    instance itself:

    >>> he.exited_with_exception
    ZeroDivisionError('division by zero')

    If all you want to do though is print a string (and have the same string
    available in the `exit_value` attribute, we got you covered!
    Just specify a string and we'll make that printer callaback for you!

    >>> from functools import partial
    >>>
    >>> with HandleExceptions({ZeroDivisionError: "You interrupted me again!"}):
    ...     print('This also works')
    ...     0 / 0
    This also works
    You interrupted me again!

    Note that specifying `partial(print, "some message")` will work as a
    "printing callback", but the string won't be available in `.exit_value` since
    `print` returns None.

    A few recipes now...

    You can also use your own custom exception types to do things like interrupt
    a with block early given some condition(s).

    >>> with HandleExceptions(
    ...     {InterruptWithBlock: "The with block was interrupted early."}
    ... ):
    ...     print('before condition')
    ...     x = 5 % 2
    ...     if x:
    ...         raise InterruptWithBlock()
    ...     print('after condition')
    ...
    ...
    before condition
    The with block was interrupted early.

    Tip: If you need to do stuff with an exception, but reraise it, you can
    still do that in your callback. Just say `raise` at the end of the callback!

    >>> def print_and_raise(msg):
    ...     print(msg)
    ...     raise
    >>>
    >>> with HandleExceptions({  # doctest: +SKIP
    ...     ZeroDivisionError: partial(print_and_raise, "That again!"),
    ... }):
    ...     print('This also works')
    ...     0 / 0
    This also works
    That again!
    Traceback (most recent call last):
        ...
    ZeroDivisionError: division by zero
    """

    on_error: Mapping[ExceptionType, OnErrorCallback] = dataclasses.field(
        default_factory=dict
    )
    exited_with_exception = None

    def __post_init__(self):
        self.on_error = dict(self.on_error)
        for handled_exc_type, callback in self.on_error.items():
            if isinstance(callback, str):
                msg = callback
                self.on_error[handled_exc_type] = partial(log_and_return, msg)

    def __enter__(self):
        self.initialize()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.exited_with_exception = exc_val
            for handled_exc_type, callback in self.on_error.items():
                if issubclass(exc_type, handled_exc_type):
                    # if the exc_type is a subtype of handled_exc_type, call the callback
                    self.exit_value = callback()  # storing the result in exit_value
                    # and exit with True
                    return True  # suppress the exception
            # if there was an exception, but a handled one, raise it!
            raise

    def exited_with_handled_exception(self):
        return hasattr(self, 'exit_value')

    def initialize(self):
        if hasattr(self, 'exit_value'):
            delattr(self, 'exit_value')
        self.exited_with_exception = None



"""Scrap"""
from functools import wraps
from typing import Iterable, Callable, Tuple
from i2.wrapper import Wrap
from i2 import call_forgivingly, Sig

# NOTE: Wrapx is now in i2.wrapper and test in i2.tests.wrapper_test


def default_caller(
    func: Callable, func_args: tuple, func_kwargs: dict, **caller_params
):
    """Function is called normally. caller_kwargs ignored"""
    return func(*func_args, **func_kwargs)


def _extract_params(ingress_output) -> Tuple[tuple, dict, dict, dict]:
    """Pads tuple with empty dicts"""
    if isinstance(ingress_output, tuple) and len(ingress_output) == 2:
        # This is the "normal" ingress protocol, so transform to the extended
        return ingress_output[0], ingress_output[1], {}, {}
    elif isinstance(ingress_output, dict):
        # This is the extended ingress protocol
        return (
            ingress_output.get('inner_args', ()),
            ingress_output.get('inner_kwargs', {}),
            ingress_output.get('caller_kwargs', {}),
            ingress_output.get('egress_kwargs', {}),
        )
        # TODO: Consider making a ingress_output object to avoid key-typo errors
        # TODO: Should we assert those were the only keys to mitigate key-typo errors?
    else:
        raise ValueError(
            f'ingress_output should be a 2-tuple or a dict. Was {ingress_output}'
        )


# TODO: Create a few higher-level constructors
# TODO: Add static analysis to ascertain protocols
class Wrapx1(Wrap):
    """An extended ``Wrap`` object that enables control over ``caller`` & ``egress``.

    To get the extended functionality, the ``ingress`` function must return a dict with
    keys ``inner_args, inner_kwargs, caller_kwargs`` and/or  ``egress_kwargs``.
    If not specified, the ``inner_args`` value will default to the empty tuple ``()``,
    and all others to the empty dictionary ``{}``.

    The additional control arguments be keyword-only in ``ingress`` as well as in the
    ``caller`` and/or ``egress`` they control.

    >>> from inspect import signature
    >>> def func(x, y):
    ...     return x + y

    Let's wrap the function endowing the wrapped function with an extra argument, ``z``,
    which can be used to multiply the output.

    >>> def ingress(x, y, *, z):
    ...     return dict(inner_args=(x, y), egress_kwargs=dict(z=z))
    >>> def egress(v, *, z):
    ...     return v * z
    ...
    >>> wrapped_func = Wrapx1(func, ingress=ingress, egress=egress)
    >>>
    >>> func(1, 2)
    3
    >>> str(signature(wrapped_func))
    '(x, y, *, z)'
    >>> assert wrapped_func(1, 2, z=3) == 9 == func(1, 2) * 3

    Here's a bit more realistic application. Say we want to add two arguments,
    ``s`` (a mapping) and ``k`` (a key) that will have the effect of writing outputs
    of the function in ``s[k]``.

    >>> def _saving_ingress(x, y, *, k, s):
    ...     return dict(inner_args=(x, y), egress_kwargs=dict(k=k, s=s))
    >>> def _saving_egress(v, *, k, s):
    ...     s[k] = v
    ...     return v
    >>> save_on_output = Wrapx1(func, ingress=_saving_ingress, egress=_saving_egress)
    >>> str(signature(save_on_output))
    '(x, y, *, k, s)'
    >>>
    >>> store = dict()
    >>> assert save_on_output(1, 2, k='save_here', s=store) == 3 == func(1, 2)
    >>> store
    {'save_here': 3}

    An graph exhibiting how it works:

    .. image:: https://user-images.githubusercontent.com/1906276/159377430-e67f4fb7-22f2-44cd-97f8-6ddbbe53f66d.png


    """

    def __init__(
        self, func, ingress=None, egress=None, *, name=None, caller=default_caller
    ):
        super().__init__(func, ingress=ingress, egress=egress, name=name)
        self.caller = caller

    def __call__(self, *outer_args, **outer_kwargs):
        func_args, func_kwargs, caller_params, egress_params = _extract_params(
            self.ingress(*outer_args, **outer_kwargs)
        )
        # call the function
        func_output = self.caller(self.func, func_args, func_kwargs, **caller_params)
        # process output with egress
        return self.egress(func_output, **egress_params)


def _pad_with_empty_dicts(tup, target_length=4):
    """Pads tuple with empty dicts"""
    return tup + tuple([{}] * max(0, target_length - len(tup)))


# TODO: Add static analysis to ascertain dynamic _pad_with_empty_dicts won't lead to
#  trouble
class Wrapx2(Wrap):
    """
    >>> from inspect import signature
    >>> def func(x, y):
    ...     return x + y
    >>> def ingress(x, y, *, z):
    ...     return (x, y), {}, dict(z=z), {}
    >>> def egress(v, *, z):
    ...     return v * z
    >>> wrapped_func = Wrapx2(func, ingress=ingress, egress=egress)
    >>>
    >>> assert func(1, 2) == 3
    >>> assert str(signature(wrapped_func)) == '(x, y, *, z)'
    >>> assert wrapped_func(1, 2, z=3) == 9 == func(1, 2) * 3
    """

    def __init__(
        self, func, ingress=None, egress=None, *, name=None, caller=default_caller
    ):
        super().__init__(func, ingress=ingress, egress=egress, name=name)
        self.caller = caller

    def __call__(self, *outer_args, **outer_kwargs):
        func_args, func_kwargs, egress_params, caller_params = _pad_with_empty_dicts(
            self.ingress(*outer_args, **outer_kwargs)
        )
        # call the function
        func_output = self.caller(self.func, func_args, func_kwargs, **caller_params)
        # process output with egress
        return self.egress(func_output, **egress_params)


# from i2.deco import double_up_as_factory
#
# # TODO: Other design options? Example making an Egress class that has extract_kwargs meth
# # TODO: Make "copy" of function before adding extract_kwargs attr, or let user do it
# #  if it matters? (I vote let user)
# # TODO: Add ability to not pop extracted
# # TODO: Postelize extract_kwargs further?
# @double_up_as_factory
# def add_extract_kwargs(func=None, *, extract_kwargs):
#     if isinstance(extract_kwargs, Iterable):
#         keys_to_extract = tuple(extract_kwargs)
#         extract_kwargs = partial(_extract_kwargs, keys_to_extract=keys_to_extract)
#     if callable(extract_kwargs) and Sig(extract_kwargs).n_required <= 1:
#         setattr(func, 'extract_kwargs', extract_kwargs)
#     else:
#         setattr(func, 'extract_kwargs', _return_empty_dict)
#
#     return func
#
#
# def _return_empty_dict(kwargs):
#     """Used when we want to have an 'empty' extract_kwargs on an egress or caller"""
#     return dict()
#
#
# def _extract_kwargs(kwargs, keys_to_extract):
#     return {k: kwargs.pop(k) for k in keys_to_extract}
#
#
# @add_extract_kwargs(extract_kwargs=_return_empty_dict)
# def default_caller(
#     func: Callable, func_args: tuple, func_kwargs: dict, caller_kwargs
# ):
#     """Function is called normally. caller_kwargs ignored"""
#     return func(*func_args, **func_kwargs)
#
#
# # TODO: Better have an egress that always has a (potentially empty) extract_kwargs or
# #  check for existence at runtime?
# class Wrapxx(Wrap):
#     def __init__(
#             self, func, ingress=None, egress=None, *, name=None, caller=default_caller
#     ):
#         super().__init__(func, ingress=ingress, egress=egress, name=name)
#         self.caller = caller
#         # current_sig = Sig(self)
#         # # extended_sig = Sig(self) +
#         # self.__signature__ = Sig(
#         #     extended_sig, return_annotation=current_sig.return_annotation
#         # )
#
#     def __call__(self, *user_args, **user_kwargs):
#         # extract the kwargs the egress needs from user input
#         egress_kwargs = self.egress.extract_kwargs(user_kwargs)
#         # extract the kwargs the caller needs from user input
#         caller_kwargs = self.caller.extract_kwargs(user_kwargs)
#         # prepare function inputs
#         func_args, func_kwargs = self.ingress(*user_args, **user_kwargs)
#         # call the function
#         func_output = self.caller(self.func, func_args, func_kwargs, caller_kwargs)
#         # process output with egress
#         return self.egress(func_output, **egress_kwargs)
#



"""
# Signature Calculus

`Sig` is an extension of the `inspect.Signature` object that puts more
goodies at your fingertips.

First of all, a `Sig` instance can be made from a variety of types


>>> from i2.signatures import *
>>> Sig()
<Sig ()>
>>> Sig('self')
<Sig (self)>
>>> Sig(lambda a, b, c=0: None)
<Sig (a, b, c=0)>

>>> Sig(Parameter('foo', Parameter.POSITIONAL_ONLY, default='foo', annotation=int))
<Sig (foo: int = 'foo', /)>
>>> Sig() + 'self' + (lambda a, b, c=0: None) - 'c' + P('c', default=1)
<Sig (self, a, b, c=1)>


Note a difference between `Sig.extract_args_and_kwargs` and binding by hand.

>>> def formula1(w, /, x: float, y=1, *, z: int = 1):
...     return ((w + x) * y) ** z
...
>>>
>>> sig = Sig(formula1)
>>> b = sig.bind(1,2,y=10)
>>> b.apply_defaults()
>>> b.args, b.kwargs
((1, 2, 10), {'z': 1})
>>>
>>> args, kwargs = sig.extract_args_and_kwargs(1, 2, y=10, _apply_defaults=True)
>>> assert args == (1,)
>>> assert kwargs == {'x': 2, 'y': 10, 'z': 1}

bind method seems to favor args over kwargs.

"""



"""Simple explorations of doc mints"""
import inspect
from collections import defaultdict


class NoDefault(object):
    def __repr__(self):
        return 'no_default'


no_default = NoDefault()

NO_NAME = '_no_name'

valid_types = [str, dict, list, float, int, bool]

types_map = {
    str: 'string',
    dict: 'object',
    list: 'array',
    float: 'float',
    int: 'int',
    bool: 'boolean',
    '{}': '{}',
    None: '{}',
}


def name_of_obj(o):
    if hasattr(o, '__name__'):
        return o.__name__
    elif hasattr(o, '__class__'):
        return name_of_obj(o.__class__)
    else:
        return NO_NAME


def parse_mint_doc(doc: str) -> dict:
    if doc is None:
        doc = ''
    DESCRIPTION = 0
    PARAMS = 1
    RETURN = 2
    split_doc = doc.split('\n')
    summary = ''
    description_lines = []
    inputs = defaultdict(dict)
    return_value = {}
    tags = []
    reading = DESCRIPTION
    cur_item_name = ''
    for line in split_doc:
        line = line.strip()
        if line.startswith(':param'):
            reading = PARAMS
            split_line = line.split(':')
            param_def = split_line[1]
            split_param_def = param_def.split(' ')
            param_name = ''
            param_type = None
            if len(split_param_def) >= 3:
                param_type = split_param_def[1]
                param_name = split_param_def[2]
                try:
                    assert len(param_type) <= 5
                    param_type = eval(param_type)
                    assert param_type in valid_types
                except Exception:
                    param_type = '{}'
            elif len(split_param_def) == 2:
                param_name = split_param_def[1]
            if param_name:
                cur_item_name = param_name
                if param_type is not None:
                    inputs[param_name] = {'type': param_type}
                if param_name in inputs:
                    inputs[param_name]['description'] = (
                        ':'.join(split_line[2:]) if len(split_line) > 2 else ''
                    )
        elif line.startswith(':return'):
            reading = RETURN
            split_line = line.split(':')
            return_type = split_line[1]
            try:
                assert len(return_type) <= 5
                return_type = eval(return_type)
                assert return_type in valid_types
            except Exception:
                return_type = '{}'
            return_value = {'type': return_type}
            return_value['description'] = (
                ':'.join(split_line[2:]) if len(split_line) > 2 else ''
            )
        elif line.startswith(':tags'):
            split_line = line.split(':')
            tags = split_line[1].replace(' ', '').split(',')
        else:
            if reading == DESCRIPTION:
                if not summary:
                    summary = line
                else:
                    description_lines.append(line)
            elif reading == PARAMS:
                if 'description' not in inputs[cur_item_name]:
                    inputs[cur_item_name]['description'] = ''
                inputs[cur_item_name]['description'] += ' ' + line
            elif reading == RETURN:
                return_value['description'] += ' ' + line
    return {
        'description': ' '.join(description_lines),
        'inputs': dict(inputs),
        'return': return_value,
        'summary': summary,
        'tags': tags,
    }


# TODO: Expand so that user can specify what to include in the mint
def mint_of_callable(f):
    """
    Get meta-data about a callable.
    :param f: A callable (function, method, ...)
    :return: A dict containing information about the interface of f, that is, name, module, doc, and
    input and output information.

    >>> def test_callable(arg1, arg2: int) -> str:
    ...     __doc__ = "A testable callable\\n:param str arg1: A string\\n"
    ...     print('calling!')
    ...     print(arg1)
    ...     print(arg2)
    ...     return 'returned'
    ...
    >>> mint = mint_of_callable(test_callable)
    >>> list(mint.keys())
    ['name', 'module', 'doc', 'description', 'summary', 'tags', 'input', 'output']
    >>> mint['name']
    'test_callable'
    >>> assert mint['module'] == test_callable.__module__, "mint['module'] == test_callable.__module__"
    >>> mint['input']
    {'arg1': {}, 'arg2': {'annot': <class 'int'>}}
    >>> mint['output']
    {'annot': <class 'str'>}
    >>> assert mint.pop('module').endswith('simple_pymint')  # popping off because module is relative so not consistent
    >>> from pprint import pprint
    >>> pprint(mint)
    {'description': '',
     'doc': None,
     'input': {'arg1': {}, 'arg2': {'annot': <class 'int'>}},
     'name': 'test_callable',
     'output': {'annot': <class 'str'>},
     'summary': '',
     'tags': []}
    """
    raw_doc = inspect.getdoc(f)
    parsed_doc = parse_mint_doc(raw_doc)
    doc_inputs = parsed_doc['inputs']
    doc_return = parsed_doc['return']
    mint = {
        'name': name_of_obj(f),  # TODO: Better NO_NAME or just not the name field?
        'module': f.__module__,
        'doc': raw_doc,
        'description': parsed_doc['description'] or '',
        'summary': parsed_doc['summary'],
        'tags': parsed_doc['tags'],
    }

    argspec = inspect.getfullargspec(f)
    annotations = argspec.annotations
    input_specs = {}
    args = argspec.args or []
    defaults = argspec.defaults or []
    for arg_name, dflt in zip(
        args, [no_default] * (len(args) - len(defaults)) + list(defaults)
    ):
        input_specs[arg_name] = {}
        if dflt is not no_default:
            input_specs[arg_name]['default'] = dflt
        if arg_name in doc_inputs:
            doc_input_arg = doc_inputs[arg_name]
            input_specs[arg_name]['doc_type'] = doc_input_arg['type']
            input_specs[arg_name]['description'] = doc_input_arg['description']
        if arg_name in annotations:
            input_specs[arg_name]['annot'] = annotations[arg_name]
        # if input_specs[arg_name].get('type', None):  # NOTE: Note the pymint concern to convert to string
        #     input_specs[arg_name]['type'] = types_map.get(input_specs[arg_name]['type'], '{}')

    mint['input'] = input_specs

    mint['output'] = {}
    if doc_return:
        mint['output'] = doc_return
    if 'return' in annotations:
        mint['output']['annot'] = annotations['return']

    return mint


""" TODO:
    - indicate required/optional or nullable input arguments
    - indicate input argument constraints (min, max, enum, etc)
    - allow listing properties of complex objects as input arguments
        (eg argument a1 type is a dict with expected properties p1, p2, p3)
"""

if __name__ == '__main__':
    from pprint import pprint

    def _test_callable(arg1, arg2: int) -> str:
        """
        A testable callable.
        :param str arg1: A string
        :params float arg2: A number
        """
        print('calling!')
        print(arg1)
        print(arg2)
        return 'returned'

    minted = mint_of_callable(_test_callable)
    print('mint of _test_callable')
    pprint(minted)

    # and a bit of naughty...

    print('mint of pprint')
    pprint(mint_of_callable(pprint))

    print('mint of mint_of_callable')
    pprint(mint_of_callable(mint_of_callable))



"""
Tools to make input and output transforming decorators.

Input value transformers can be conditioned on argument value and name, as well as the
wrapped function itself.

Output value tranformers can be conditioned on argument value and the wrapped function.

"""
from dataclasses import dataclass
from typing import (
    Mapping,
    Callable,
    Optional,
    TypedDict,
)
from inspect import signature, Parameter
from pickle import dumps

from i2.signatures import Sig

import functools

# from typing import _TypedDictMeta
# Note: When doing the above, lint complains, but TypedDict alone doesn't do the trick.
# _TypedDictMeta = TypedDict  # to show that TypedDict doesn't work
# Raises     TypeError: TypedDict does not support instance and class checks
# So we do the following hack:
_TypedDictMeta = type(TypedDict)


# monkey patching WRAPPER_ASSIGNMENTS to get "proper" wrapping (adding defaults and kwdefaults
wrapper_assignments = (
    '__module__',
    '__name__',
    '__qualname__',
    '__doc__',
    '__annotations__',
    '__defaults__',
    '__kwdefaults__',
)

update_wrapper = functools.update_wrapper
update_wrapper.__defaults__ = (
    functools.WRAPPER_ASSIGNMENTS,
    functools.WRAPPER_UPDATES,
)
wraps = functools.wraps
wraps.__defaults__ = (functools.WRAPPER_ASSIGNMENTS, functools.WRAPPER_UPDATES)


def identity_func(x):
    return x


@dataclass
class IoTrans:
    def in_val_trans(self, argval, argname, func):
        return argval  # default is the value as is

    def out_trans(self, argval, func):
        return argval  # default is the value as is

    def __call__(self, func):
        sig = Sig(func)

        @wraps(
            func
        )  # Todo: Want empty mapping as default (use frozendict or __post_init__?)
        def wrapped_func(*args, **kwargs):
            original_kwargs = sig.extract_kwargs(*args, **kwargs)
            new_kwargs = {
                argname: self.in_val_trans(argval, argname, func)
                for argname, argval in original_kwargs.items()
            }
            new_args, new_kwargs = sig.args_and_kwargs_from_kwargs(new_kwargs)
            return self.out_trans(func(*new_args, **new_kwargs), func)

        return wrapped_func


@dataclass
class ArgnameIoTrans(IoTrans):
    """Transforms argument values according to their names

    >>> def foo(a, b=1.0):
    ...     return a + b
    >>>
    >>> input_trans = ArgnameIoTrans({'a': int, 'b': float})
    >>> foo2 = input_trans(foo)
    >>> assert foo2(3) == 4.0
    >>> assert foo2(-2, 2) == 0.0
    >>> assert foo2("3") == 4.0
    >>> assert foo2("-2", "2") == 0.0
    >>> assert signature(foo) == signature(foo2)
    """

    argname_2_trans_func: Mapping

    def in_val_trans(self, argval, argname, func):
        trans_func = self.argname_2_trans_func.get(argname, None)
        if trans_func is not None:
            return trans_func(argval)
        else:
            return super().in_val_trans(argval, argname, func)


empty = Parameter.empty


@dataclass
class AnnotAndDfltIoTrans(IoTrans):
    """Transforms argument values using annotations and default type

    >>> def foo(a: int, b=1.0):
    ...     return a + b
    >>>
    >>> input_trans = AnnotAndDfltIoTrans()
    >>> foo3 = input_trans(foo)
    >>> assert foo3(3) == 4.0
    >>> assert foo3(-2, 2) == 0.0
    >>> assert foo3("3") == 4.0
    >>> assert foo3("-2", "2") == 0.0
    >>> assert signature(foo) == signature(foo3)
    """

    annotations_handled = frozenset([int, float, str])
    dflt_types_handled = frozenset([int, float, str])

    def in_val_trans(self, argval, argname, func):
        param = signature(func).parameters[argname]
        if param.annotation in self.annotations_handled:
            return param.annotation(argval)
        else:
            dflt_type = type(param.default)
            if dflt_type in self.dflt_types_handled:
                return dflt_type(argval)
        return super().in_val_trans(argval, argname, func)


@dataclass
class JSONAnnotAndDfltIoTrans(AnnotAndDfltIoTrans):
    """Transforms argument values using annotations and default type,
    including lists, iterables, dicts, and booleans

    >>> def foo(a: dict, b=['dflt'], c=False):
    ...     return dict({}, a=a, b=b, c=c)
    >>>
    >>> input_trans = JSONAnnotAndDfltIoTrans()
    >>> foo4 = input_trans(foo)
    >>> assert foo4('{}') == {'a': {}, 'b': ['dflt'], 'c': False}
    >>> assert foo4({'d': 'e'}, '["some_value"]', 'true') == {'a': {'d': 'e'}, 'b': ['some_value'], 'c': True}
    >>> complex_types_result = foo4('{"None": null, "True": true, "False": false}', '[null, true, false]', 'false')
    >>> assert complex_types_result == {'a': {'None': None, 'True': True, 'False': False}, 'b': [None, True, False], 'c': False}
    >>> assert signature(foo) == signature(foo4)
    """

    def in_val_trans(self, argval, argname, func):
        param = signature(func).parameters[argname]
        if param.annotation != str and not isinstance(param.default, str):
            if (
                param.annotation == dict
                or isinstance(param.annotation, _TypedDictMeta)
                or isinstance(param.default, dict)
                or isinstance(type(param.default), _TypedDictMeta)
                or param.annotation == bool
                or isinstance(param.default, bool)
            ):
                return cast_to_jdict(argval)
            if hasattr(param.annotation, '__iter__') or hasattr(
                param.default, '__iter__'
            ):
                return cast_to_list(argval)
        return super().in_val_trans(argval, argname, func)


# TODO: Move this doctest to tests suite that ignores the test if pandas not there!
# >>> import pandas as pd
# >>> out_trans = TypedBasedOutIoTrans({
# ...     (list, tuple, set): ', '.join,
# ...     pd.DataFrame: pd.DataFrame.to_csv
# ... })
# >>>
# >>>
# >>> @out_trans
# ... def repeat(a: int, b: list):
# ...     return a * b
# ...
# >>> assert repeat(2, ['repeat', 'it']) == 'repeat, it, repeat, it'
# >>>
# >>> @out_trans
# ... def transpose(df):
# ...     return df.T
# ...
# >>> df = pd.DataFrame({'a': [1,2,3], 'b': [10, 20, 30]})
# >>> print(df.to_csv())  # doctest: +NORMALIZE_WHITESPACE
# ,a,b
# 0,1,10
# 1,2,20
# 2,3,30
# >>> print(transpose(df))  # doctest: +NORMALIZE_WHITESPACE
# ,0,1,2
# a,1,2,3
# b,10,20,30
# TODO: Write doctest (that only needs builtins)
@dataclass
class TypedBasedOutIoTrans(IoTrans):
    """Transform output according to it's type."""

    trans_func_for_type: Mapping = ()  # Todo: Want empty mapping as default (use frozendict or __post_init__?)
    dflt_trans_func: Optional[Callable] = None

    def out_trans(self, argval, func):
        for typ in self.trans_func_for_type:
            if isinstance(argval, typ):
                return self.trans_func_for_type[typ](argval)
        if isinstance(
            self.dflt_trans_func, Callable
        ):  # Question: use callable() instead? What's the difference?
            return self.dflt_trans_func(argval)


def pickle_out_trans(self, argval, func):
    return dumps(argval)


PickleFallbackTypedBasedOutIoTrans = functools.partial(
    TypedBasedOutIoTrans, dflt_trans_func=dumps
)

import json
import os


def cast_to_jdict(value):
    """Tries to cast to a json-friendly dictionary.

    >>> cast_to_jdict('3')
    [3]
    >>> cast_to_jdict("[3]")
    [3]
    >>> cast_to_jdict("[4,2]")
    [4, 2]
    >>> cast_to_jdict('[4, "string", ["another", "list"], {"nested": 10.2}]')
    [4, 'string', ['another', 'list'], {'nested': 10.2}]
    >>> cast_to_jdict('{"here": "is", "a": {"nested": "json"}, "with": [null, true, false, 1, 2.3]}')
    {'here': 'is', 'a': {'nested': 'json'}, 'with': [None, True, False, 1, 2.3]}

    And csvs too:

    >>> cast_to_jdict('1,2,3.4, "string" ,  null, true, false, ["a", "list"]')
    [1, 2, 3.4, 'string', None, True, False, ['a', 'list']]
    """
    if isinstance(value, str):
        value = value.strip()
        if value:
            first_char = value[0]
            if first_char in {'[', '{'}:
                return json.loads(value)
            elif value in ['true', 'false', 'null']:
                return json.loads(value)
            elif os.path.isfile(value):
                return json.load(value)
            else:
                return json.loads(
                    '[' + value + ']'
                )  # wrap in brackets and call json.loads
        else:
            return ''
    else:
        return value


def cast_to_list(value):
    """Tries to case to a list (with json friendly elements)

    >>> cast_to_list('3')
    [3]
    >>> cast_to_list("[3]")
    [3]
    >>> cast_to_list("[4,2]")
    [4, 2]
    >>> cast_to_list('[4, "string", ["another", "list"], {"nested": 10.2}]')
    [4, 'string', ['another', 'list'], {'nested': 10.2}]

    And csvs too:

    >>> cast_to_list('1,2,3.4, "string" ,  null, true, false, ["a", "list"]')
    [1, 2, 3.4, 'string', None, True, False, ['a', 'list']]
    """
    if isinstance(value, str):
        value = cast_to_jdict(value)
        assert isinstance(value, list)
        return value
    elif hasattr(value, 'tolist'):  # meant for numpy arrays
        # what other potential attributes to check for?
        return value.tolist()
    else:
        return list(
            value
        )  # will work with set, tuple, and other iterables (not recursively though: just level 0)


# @dataclass
# class PickleFallbackTypedBasedOutIoTrans(TypedBasedOutIoTrans):
#     dflt_trans_func = pickle_out_trans



from i2 import Sig
from i2.signatures import _remove_variadics_from_sig, ch_variadics_to_non_variadic_kind
from i2.wrapper import InnerMapIngress
from i2.wrapper import wrap
from inspect import Parameter, Signature


# def bar(w, /, x: float, y=1, *args, z: int = 1, **rest):
#     return ((w + x) * y) ** z


# bar_sig = Sig(bar)


def foo(po, *vp, ko, **vk):
    return f'{po=}, {vp=}, {ko=}, {vk=}'


new_sig = _remove_variadics_from_sig(Sig(foo))
assert str(new_sig) == '(po, vp=(), *, ko, vk={})'

ingress = InnerMapIngress.from_signature(foo, outer_sig=new_sig)

if __name__ == '__main__':
    new_foo = wrap(foo, ingress=ingress)
    # ingress_args = (1, (2, 3))
    # ingress_kwargs = {"ko": 4, "vk": {"hello": "world"}}
    # print(ingress(*ingress_args, **ingress_kwargs))

    assert (
        foo(1, 2, 3, ko=4, hello='world')
        == "po=1, vp=(2, 3), ko=4, vk={'hello': 'world'}"
    )
    new_foo = wrap(foo, ingress=ingress)
    assert str(Sig(new_foo)) == '(po, vp=(), *, ko, vk={})'
    # assert (
    #     new_foo(1, (2, 3), ko=4, vk=dict(hello="world"))
    #     == "po=1, vp=(2, 3), ko=4, vk={'hello': 'world'}"
    # )
    # args, kwargs = foo_sig.args_and_kwargs_from_kwargs(dict(w=4, x=3, y=2, z=1, t=12))

    # print(f"args:{args}, kwargs: {kwargs}")
    # print(new_foo(1, (2, 3), ko=4, vk=dict(hello="world")))
